#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

import unittest

from pyspark.sql.functions import udf
from pyspark.sql.tests.test_udf import BaseUDFTestsMixin
from pyspark.sql.types import Row, StringType
from pyspark.testing.sqlutils import (
    ReusedSQLTestCase,
)
import numpy
from datetime import datetime


class SafeSparkPythonUDFTests(BaseUDFTestsMixin, ReusedSQLTestCase):
    @classmethod
    def setUpClass(cls):
        super(SafeSparkPythonUDFTests, cls).setUpClass()
        # Ensure the isolated udf is enabled
        assert cls.spark.conf.get("spark.databricks.pyspark.isolatedUDF.enabled")

    # Test cases skipped as in pyspark.databricks.sql.tests.test_arrow_python_udf.
    @unittest.skip("Unrelated test, and it fails when it runs duplicatedly.")
    def test_register_java_function(self):
        super(SafeSparkPythonUDFTests, self).test_register_java_function()

    @unittest.skip("Unrelated test, and it fails when it runs duplicatedly.")
    def test_register_java_udaf(self):
        super(SafeSparkPythonUDFTests, self).test_register_java_udaf()

    # The tests skipped below are limitations / TODOs for safe spark.
    @unittest.skip("Broadcast not yet supported in safe spark udf.")
    def test_broadcast_in_udf(self):
        super(SafeSparkPythonUDFTests, self).test_broadcast_in_udf()

    @unittest.skip("Accumulator not yet supported in safe spark udf.")
    def test_same_accumulator_in_udfs(self):
        super(SafeSparkPythonUDFTests, self).test_same_accumulator_in_udfs()

    # TODO: add coverage for more types.
    def test_udf_types(self):
        def typeof_recursive(v):
            if isinstance(v, dict):
                res = []
                for k in v:
                    res = res + [f"{str(k)}({typeof_recursive(k)}):{typeof_recursive(v[k])}"]
                return "dict{" + ",".join(res) + "}"
            elif isinstance(v, list):
                return "list[" + typeof_recursive(v[0]) + "]"
            elif isinstance(v, Row):
                res = []
                for k in v.asDict():
                    res = res + [str(k) + ":" + typeof_recursive(v[k])]
                return "pyspark.sql.types.Row{" + ",".join(res) + "}"
            elif isinstance(v, numpy.ndarray):
                return "numpy.ndarray[" + typeof_recursive(v[0]) + "]"
            else:
                return type(v).__name__

        @udf(returnType=StringType())
        def python_type_value(value):
            return typeof_recursive(value) + "," + str(value)

        def get_sql_type(v):
            df = self.spark.createDataFrame([Row(value=v)])
            return str(df.schema.fields[0].dataType)

        def run_udf(v):
            df = self.spark.createDataFrame([Row(value=v)])
            return df.select(python_type_value("value")).collect()[0][0]

        def objFromSql(sql: str):
            return self.spark.sql(f"select {sql} as value").collect()[0][0]

        named_struct_sql = "NAMED_STRUCT('a', 1, 'b', 'hello_named_struct')"
        struct_sql = "STRUCT(1, 'hello_struct')"
        map_sql = "MAP(1, 'hello_map')"
        timestamp_sql = "TIMESTAMP'2020-02-02 12:15:16.123'"

        class Base:
            def __init__(self):
                self.base = 1

        class Child:
            def __init__(self):
                self.child = Base()

        timestamp = datetime.fromisoformat("2022-01-01 12:05:23.283")

        testcases = [
            # objects from Python
            (
                {"b": "hello_pydict", "a": 1},
                "MapType(StringType(), StringType(), True)",
                "dict{a(str):str,b(str):str}",
                {"a": "1", "b": "hello_pydict"},
            ),
            (
                [1, 2, 3],
                "ArrayType(LongType(), True)",
                "list[int]",
                [1, 2, 3],
            ),
            # objects from SQL
            (
                objFromSql(named_struct_sql),
                "StructType([StructField('a', LongType(), True), StructField('b', StringType(), True)])",  # noqa: E501
                "pyspark.sql.types.Row{a:int,b:str}",
                Row(a=1, b="hello_named_struct"),
            ),
            (
                objFromSql(struct_sql),
                "StructType([StructField('col1', LongType(), True), StructField('col2', StringType(), True)])",  # noqa: E501
                "pyspark.sql.types.Row{col1:int,col2:str}",
                Row(col1=1, col2="hello_struct"),
            ),
            (
                objFromSql(map_sql),
                "MapType(LongType(), StringType(), True)",
                "dict{1(int):str}",
                {1: "hello_map"},
            ),
            # nested
            (
                [[1.0, 2.0], [2.0, 3.0]],
                "ArrayType(ArrayType(DoubleType(), True), True)",
                "list[list[float]]",
                [[1.0, 2.0], [2.0, 3.0]],
            ),
            (
                [Base(), Base()],
                "ArrayType(StructType([StructField('base', LongType(), True)]), True)",
                "list[pyspark.sql.types.Row{base:int}]",
                [Row(base=1), Row(base=1)],
            ),
            # The following test cases still show inconsistent results compared to Pyspark.
            # Note that this is not a regression caused by the type conversion logics, but due to
            # the pyarrow to pandas library.
            # TODO(SAS-815): further conversion get closer to full consistensy.
            (
                {"a": 1, "b": "hello"},
                "MapType(StringType(), LongType(), True)",
                # We got this float due to pyarrow.Table.to_pandas().
                # TODO: decide whether or not to apply explicitly cast for such cases.
                "dict{a(str):int,b(str):NoneType}",
                # The string value got type coercion.
                {"a": 1, "b": None},
            ),
            (
                [1, 2, 3, None],
                "ArrayType(LongType(), True)",
                "list[int]",
                [1, 2, 3, None],
            ),
            (
                objFromSql(timestamp_sql),
                "TimestampType()",
                "datetime",
                # Pyspark's result 2020-02-02 12:15:16.123000
                # Arrow->pandas's timestamp is more accurate than python.
                "2020-02-02 12:15:16.123000+00:00",
            ),
            (
                {timestamp: timestamp},
                "MapType(TimestampType(), TimestampType(), True)",
                "dict{2022-01-01 12:05:23.283000+00:00(datetime):datetime}",
                "{datetime.datetime(2022, 1, 1, 12, 5, 23, 283000, "
                # in pyspark, the tzinfo is not shown
                + "tzinfo=<StaticTzInfo 'Etc/UTC'>)"
                + ": datetime.datetime(2022, 1, 1, 12, 5, 23, 283000, "
                + "tzinfo=<StaticTzInfo 'Etc/UTC'>)}"
                # str({timestamp: timestamp}),
            ),
        ]

        # There are some types not supported in pyarrow.to_pandas library.
        # It is not a regression caused by the type conversion logics.
        # For example: Map in an array type.
        # (
        #    [{"a": 1.0}, {"b": 2.0}],
        #    "ArrayType(MapType(StringType(), DoubleType(), True), True)",
        #    "list[dict{a: float, b: float}]",
        #    [{"a": 1.0}, {"b": 2.0}],
        # ),

        for (v, expected_sql_type, expected_udf_type, expected_udf_value) in testcases:
            result = run_udf(v)
            self.assertEquals(get_sql_type(v), expected_sql_type)
            self.assertEquals(result, expected_udf_type + "," + str(expected_udf_value))


if __name__ == "__main__":
    from pyspark.databricks.sql.tests.safespark.test_safespark_python_udf import *  # noqa: E501,F401,F403

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
