#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

import os
import time
import unittest
import shutil
import tempfile

from pyspark.sql import DataFrame, Row
from pyspark.sql.functions import col, count, lit
from pyspark.testing.sqlutils import ReusedSQLTestCase
from pyspark.sql.utils import StreamingQueryException
from pyspark.sql.types import StructType, StructField, StringType


class MetadataQueryFuzzTests(ReusedSQLTestCase):
    def setUp(self):
        # TODO: Make sure SafeFileSystem should be disabled: SC-76036
        self.spark._jvm.org.apache.spark.api.python.testutil.SafeFileSystem.disableChecking()
        self.tmp_dir = tempfile.mkdtemp(prefix=MetadataQueryFuzzTests.__name__)

    def tearDown(self):
        # TODO: Make sure SafeFileSystem should be disabled: SC-76036
        self.spark._jvm.org.apache.spark.api.python.testutil.SafeFileSystem.enableChecking()
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    @classmethod
    def setUpClass(cls):
        ReusedSQLTestCase.setUpClass()
        seed = os.getenv("FUZZ_TEST_SEED", int(round(time.time() * 1000)))
        print("using seed %s" % seed)
        cls.seed = seed
        cls.NUM_ITERATIONS = 3

    def assertEq(self, given, expected):
        self.assertEqual(
            len(given),
            len(expected),
            "The answer returned a different number of rows (%s) " % len(given)
            + "than the expected result (%s)" % (len(expected)),
        )
        for result_row in given:
            self.assertTrue(
                result_row in expected,
                "Row: %s\nnot found in result:\n%s"
                % (repr(result_row), "\n".join([repr(e) for e in expected])),
            )

    def assertContainsMetadataOptimization(self, df, is_empty):
        return (
            self.spark._jvm.__getattr__(
                "com.databricks.sql.transaction.tahoe.stats."
                + "random.DeltaMetadataQueryFuzzSuite$"
            )
            .__getattr__("MODULE$")
            .assertContainsMetadataOptimization(df, is_empty)
        )

    def assertNoMetadataOptimization(self, df, is_empty):
        return (
            self.spark._jvm.__getattr__(
                "com.databricks.sql.transaction.tahoe.stats."
                + "random.DeltaMetadataQueryFuzzSuite$"
            )
            .__getattr__("MODULE$")
            .assertNoMetadataOptimization(df, is_empty)
        )

    def run_randomized_test(self, rqg):
        rqg.generateData()
        rqg.generateQuery()

        maybe_optimized = DataFrame(rqg.planQuery(True), self.spark)
        not_optimized = DataFrame(rqg.planQuery(False), self.spark)

        self.assertEq(maybe_optimized.collect(), not_optimized.collect())

    def create_rqg(self, num_data_columns, num_partition_columns, is_empty=False):
        if is_empty:
            rows_per_file = 0
            files_per_partition = 0
        else:
            rows_per_file = 100
            files_per_partition = 10

        return (
            self.spark._jvm.__getattr__(
                "com.databricks.sql.transaction.tahoe.stats."
                + "random.DeltaMetadataQueryFuzzSuite$"
            )
            .__getattr__("MODULE$")
            .createRQG(
                self.seed,
                num_data_columns,
                num_partition_columns,
                rows_per_file,
                files_per_partition,
            )
        )

    def edge_mode_enabled(self):
        return (
            self.spark._jvm.__getattr__("com.databricks.spark.DatabricksEdgeConfigs$")
            .__getattr__("MODULE$")
            .edgeModeEnabled()
        )

    def test_partitioned_table_random(self):
        if self.edge_mode_enabled():
            for i in range(self.NUM_ITERATIONS):
                rqg = self.create_rqg(4, 1)
                self.run_randomized_test(rqg)

    def test_unpartitioned_table_random(self):
        if self.edge_mode_enabled():
            for i in range(self.NUM_ITERATIONS):
                rqg = self.create_rqg(4, 0)
                self.run_randomized_test(rqg)

    def test_empty_partitioned_table_random(self):
        if self.edge_mode_enabled():
            for i in range(self.NUM_ITERATIONS):
                rqg = self.create_rqg(4, 1, True)
                self.run_randomized_test(rqg)

    def test_empty_unpartitioned_table_random(self):
        if self.edge_mode_enabled():
            for i in range(self.NUM_ITERATIONS):
                rqg = self.create_rqg(4, 0, True)
                self.run_randomized_test(rqg)

    def test_empty_unpartitioned_table(self):
        if self.edge_mode_enabled():
            spark = self.spark
            tbl = "test_table"
            spark.sql("CREATE TABLE %s (id bigint, part string) USING delta" % tbl)
            try:
                # explicit global aggregations
                emptyAgg = {}
                self.assertEq(spark.table(tbl).agg(emptyAgg).collect(), [Row()])
                self.assertEq(spark.table(tbl).groupBy().agg(emptyAgg).collect(), [Row()])
                self.assertEq(spark.table(tbl).groupBy().agg(count("*")).collect(), [Row(0)])
                self.assertEq(spark.table(tbl).dropDuplicates().agg(emptyAgg).collect(), [Row()])
                self.assertEq(
                    spark.table(tbl).dropDuplicates().groupBy().agg(emptyAgg).collect(), [Row()]
                )
                self.assertEq(
                    spark.table(tbl).dropDuplicates().groupBy().agg(count("*")).collect(), [Row(0)]
                )

                # global aggregation is converted to grouping aggregation:
                self.assertEqual(spark.table(tbl).dropDuplicates().count(), 0)
            finally:
                spark.sql("DROP TABLE %s" % tbl)

    def test_empty_partitioned_table(self):
        if self.edge_mode_enabled():
            spark = self.spark
            tbl = "test_table"
            spark.sql(
                "CREATE TABLE %s (id bigint, part string) USING delta PARTITIONED BY (part)" % tbl
            )
            try:
                # explicit global aggregations
                emptyAgg = {}
                self.assertEq(spark.table(tbl).agg(emptyAgg).collect(), [Row()])
                self.assertEq(spark.table(tbl).groupBy().agg(emptyAgg).collect(), [Row()])
                self.assertEq(spark.table(tbl).groupBy().agg(count("*")).collect(), [Row(0)])
                self.assertEq(spark.table(tbl).dropDuplicates().agg(emptyAgg).collect(), [Row()])
                self.assertEq(
                    spark.table(tbl).dropDuplicates().groupBy().agg(emptyAgg).collect(), [Row()]
                )
                self.assertEq(
                    spark.table(tbl).dropDuplicates().groupBy().agg(count("*")).collect(), [Row(0)]
                )

                # global aggregation is converted to grouping aggregation:
                self.assertEqual(spark.table(tbl).dropDuplicates().count(), 0)
            finally:
                spark.sql("DROP TABLE %s" % tbl)

    def test_empty_partition_count(self):
        if self.edge_mode_enabled():
            spark = self.spark
            tbl = "test_table"
            spark.range(10).withColumn("part", col("id") % lit(2)).write.format(
                "delta"
            ).partitionBy("part").saveAsTable(tbl)
            try:
                query1 = "select count(*) from %s where part = 3" % tbl
                self.assertEqual(spark.sql(query1).first(), Row(0))
                query2 = "select max(id), min(id) from %s where part = 3" % tbl
                self.assertEqual(spark.sql(query2).first(), Row(None, None))
            finally:
                spark.sql("DROP TABLE %s" % tbl)

    def test_union_collect(self):
        if self.edge_mode_enabled():
            spark = self.spark
            tbl = "test_table"
            spark.range(1).write.format("delta").saveAsTable(tbl)
            try:
                query = "select distinct 'id' from %s " "union all select 'id' from %s" % (tbl, tbl)
                self.assertEqual(spark.sql(query).collect(), [Row("id"), Row("id")])
            finally:
                spark.sql("DROP TABLE %s" % tbl)

    def __checkAnswer(self, df, expectedAnswer, schema):
        expectedDF = self.spark.createDataFrame(expectedAnswer, schema)
        try:
            self.assertEqual(df.count(), expectedDF.count())
            self.assertEqual(len(df.columns), len(expectedDF.columns))
            self.assertEqual([], df.subtract(expectedDF).take(1))
            self.assertEqual([], expectedDF.subtract(df).take(1))
        except AssertionError:
            print("Expected:")
            expectedDF.show()
            print("Found:")
            df.show()
            raise

    def __create_json_file(self, json_content, file_path):
        with open(file_path, "w+") as outfile:
            outfile.write(json_content)

    def __writeToDelta(self, df, batch_id):
        base_path = "{}/test_autoloader_python_foreachbatch".format(self.tmp_dir)
        df.write.format("delta").mode("append").option("mergeSchema", "true").save(
            base_path + "/delta"
        )

    @unittest.skipIf(
        os.environ.get("PYTEST_DBCONNECT_MODE") is not None,
        "structured streaming is not supported in dbconnect",
    )
    def test_autoloader_python_foreachbatch(self):
        shutil.rmtree(self.tmp_dir, ignore_errors=True)
        base_path = "{}/test_autoloader_python_foreachbatch".format(self.tmp_dir)
        os.makedirs(base_path)

        self.spark.conf.set("spark.databricks.cloudFiles.schemaInference.sampleSize.numFiles", "1")
        self.spark.conf.set("spark.databricks.cloudFiles.rootSchemaLocation", base_path + "/schema")

        # create test json files
        os.makedirs(base_path + "/data")
        self.__create_json_file("""{"a":"0"}""", base_path + "/data/0.json")
        self.__create_json_file("""{"b":"1"}""", base_path + "/data/1.json")

        read_stream_opts = {
            "cloudFiles.format": "json",
            "cloudFiles.schemaEvolutionMode": "addNewColumns",
            "cloudFiles.maxFilesPerTrigger": "1",
        }

        # first attempt will fail with unknown fields
        with self.assertRaises(StreamingQueryException) as context:
            q = (
                self.spark.readStream.format("cloudFiles")
                .options(**read_stream_opts)
                .load(base_path + "/data")
                .writeStream.outputMode("append")
                .option("checkpointLocation", base_path + "/checkpoint")
                .foreachBatch(self.__writeToDelta)
                .queryName("test_autoloader_python_foreachbatch")
                .trigger(once=True)
                .start()
            )
            q.processAllAvailable()
        self.assertTrue("Encountered unknown field(s) during parsing:" in str(context.exception))
        q.stop()

        # restart the stream will proceed
        q = (
            self.spark.readStream.format("cloudFiles")
            .options(**read_stream_opts)
            .load(base_path + "/data")
            .writeStream.outputMode("append")
            .option("checkpointLocation", base_path + "/checkpoint")
            .foreachBatch(self.__writeToDelta)
            .queryName("test_autoloader_python_foreachbatch")
            .trigger(once=True)
            .start()
        )
        q.processAllAvailable()

        # checking if data is ingested
        schema = StructType(
            [
                StructField("a", StringType(), True),
                StructField("b", StringType(), True),
                StructField("_rescued_data", StringType(), True),
            ]
        )
        self.__checkAnswer(
            self.spark.read.format("delta").load(base_path + "/delta"),
            [("0", None, None), (None, "1", None)],
            schema,
        )
        shutil.rmtree(self.tmp_dir, ignore_errors=True)


if __name__ == "__main__":
    from pyspark.databricks.sql.tests.test_delta import *  # noqa: F401,F403

    # Some of our Python Delta tests leverage utility methods defined in test-jars
    # Here we include them through --jars
    sql_bin_dir = os.path.join(os.environ["SPARK_HOME"], "sql", "core", "target", "scala-2.12")
    test_jars = []
    if os.path.exists(sql_bin_dir):
        for f in os.listdir(sql_bin_dir):
            if "-tests.jar" in f:
                test_jars.append("--jars")
                test_jars.append(f)
        existing_args = os.environ.get("PYSPARK_SUBMIT_ARGS", "pyspark-shell")
        os.environ["PYSPARK_SUBMIT_ARGS"] = " ".join(test_jars + [existing_args])

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
