#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

import os
import shutil
import tempfile
import time
import py4j
import unittest

from pyspark.sql import SparkSession, Row
from pyspark.testing.sqlutils import ReusedSQLTestCase, have_pandas, pandas_requirement_message
from pyspark.testing.utils import ReusedPySparkTestCase


@unittest.skipIf(
    os.environ.get("PYTEST_DBCONNECT_MODE") is not None, "stream table not supported with dbconnect"
)
class SQLTests(ReusedSQLTestCase):
    # unit test of SC-10519
    def test_stream_table_options(self):
        edgeModeEnabled = (
            self.spark._jvm.__getattr__("com.databricks.spark.DatabricksEdgeConfigs$")
            .__getattr__("MODULE$")
            .edgeModeEnabled()
        )

        if not edgeModeEnabled:
            return

        for q in self.spark.streams.active:
            q.stop()
        tmpPath = tempfile.mkdtemp()
        shutil.rmtree(tmpPath)
        q = None
        tblName = "tbl"
        tempTblName = "tempTbl"
        try:
            self.spark.range(10).write.option("path", tmpPath).format("tahoe").saveAsTable(tblName)
            q = (
                self.spark.readStream.table("tbl")
                .writeStream.format("memory")
                .queryName(tempTblName)
                .start()
            )

            q.processAllAvailable()
            self.assertEqual(len(self.spark.table(tempTblName).collect()), 10)
        finally:
            if q is not None:
                q.stop()
            shutil.rmtree(tmpPath)
            self.spark.sql("drop table if exists " + tblName)
            self.spark.sql("drop view if exists " + tempTblName)

    # unit test of SC-11676
    def test_write_stream_table(self):
        edgeModeEnabled = (
            self.spark._jvm.__getattr__("com.databricks.spark.DatabricksEdgeConfigs$")
            .__getattr__("MODULE$")
            .edgeModeEnabled()
        )

        if not edgeModeEnabled:
            return

        for q in self.spark.streams.active:
            q.stop()
        tmpPath = tempfile.mkdtemp()
        shutil.rmtree(tmpPath)
        q = None
        tblName1 = "tbl_in"
        tblName2 = "tbl_out"
        try:
            self.spark.range(10).write.option("path", tmpPath).format("tahoe").saveAsTable(tblName1)
            chk = os.path.join(tmpPath, "chk")
            q = (
                self.spark.readStream.table(tblName1)
                .writeStream.format("tahoe")
                .option("checkpointLocation", chk)
                .table(tblName2)
            )

            q.processAllAvailable()
            self.assertEqual(self.spark.sql("SELECT COUNT(1) FROM " + tblName2).head()[0], 10)
        finally:
            if q is not None:
                q.stop()
            shutil.rmtree(tmpPath)
            self.spark.sql("drop table if exists " + tblName1)
            self.spark.sql("drop table if exists " + tblName2)


@unittest.skipIf(not have_pandas, pandas_requirement_message)
class TrustedFilePathTests(ReusedSQLTestCase):
    @classmethod
    def conf(cls):
        return (
            super(TrustedFilePathTests, cls)
            .conf()
            .set("spark.databricks.pyspark.trustedJVMDataTransfer.enabled", "true")
            .set("spark.databricks.pyspark.maxTrustedTempDir", 1)
            .set("spark.databricks.pyspark.trustedTempDirSecondsToLive", 2)
        )

    # Create df from a suspicious path, should get security exception
    def create_df_from_path_with_exception(self, test_path):
        import getpass

        user = getpass.getuser()
        msg = (
            "org.apache.spark.SparkSecurityException: The path %s is not in a trusted path for "
            "python user %s." % (test_path, user)
        )
        with self.assertRaisesRegexp(py4j.protocol.Py4JError, msg):
            self.spark._jsparkSession.createDataFrameFromTrustedPath(test_path, "")

    # unit test of SC-10401
    # Test feature flag effectiveness. This would first get a trusted path from jvm and run
    # createDataFrame to get another path. If the feature flag is effective, the path we get in the
    # first place would be invalid, since maxTrustedTempDir is 1.
    def test_trusted_file_effectiveness_for_df_creation(self):
        stale_dir = self.spark._jsparkSession.getTrustedPath()
        temp = tempfile.NamedTemporaryFile(dir=stale_dir, delete=False)

        # This would create a another trusted path which should evict the stale dir.
        self.spark.createDataFrame([(1, 2)])

        self.create_df_from_path_with_exception(temp.name)
        self.assertFalse(os.path.exists(stale_dir))

    # unit test of SC-10401
    # Test if the temp dir would be clean up, if timeout happens.
    def test_clean_up_when_time_out(self):
        stale_dir = self.spark._jsparkSession.getTrustedPath()
        time.sleep(5)
        # touch the cache again so that the invalidation could take place
        self.spark._jsparkSession.getTrustedPath()
        self.assertFalse(os.path.exists(stale_dir))

    # unit test of SC-10401
    # Test if the temp dir would be clean up, if error happens while reading the file.
    def test_clean_up_when_failure(self):
        stale_dir = self.spark._jsparkSession.getTrustedPath()
        temp = tempfile.NamedTemporaryFile(dir=stale_dir, delete=False)
        with self.assertRaises(py4j.protocol.Py4JError):
            # As the schema is empty, it would fail while JVM trying to parse the schema
            self.spark._jsparkSession.createDataFrameFromTrustedPath(temp.name, "")
        self.assertFalse(os.path.exists(stale_dir))

    # unit test of SC-10401
    # Test if the temp dir would be clean up, if successfully read the file.
    def test_clean_up_when_success(self):
        stale_dir = self.spark._jsparkSession.getTrustedPath()
        temp = tempfile.NamedTemporaryFile(dir=stale_dir, delete=False)
        # format a schema json so createDataFrameFromTrustedPath could be called without failure.
        schema = (
            '{"fields":[{"metadata":{},"name":"_1","nullable":true,"type":"long"}],"type":"struct"}'
        )
        self.spark._jsparkSession.createDataFrameFromTrustedPath(temp.name, schema)
        self.assertFalse(os.path.exists(stale_dir))

    # unit test of SC-10401
    # Tests for basic use cases i.e. regression
    def test_pandas_df_to_spark_df_with_trusted_path(self):
        import pandas

        pdf = pandas.DataFrame(data={"A": [1], "B": [2]})
        df = self.spark.createDataFrame(pdf)
        row = df.select("*").first()
        self.assertEqual(1, row[0])
        self.assertEqual(2, row[1])

    # unit test of SC-10401
    def test_df_from_list_with_trusted_path(self):
        lst = [(1, 2)]
        df = self.spark.createDataFrame(lst)
        row = df.select("*").first()
        self.assertEqual(1, row[0])
        self.assertEqual(2, row[1])

    # unit test of SC-10401
    # Tests trying to bypass the validation
    def test_reading_from_file_not_exists(self):
        self.create_df_from_path_with_exception("/NotExist")

    # unit test of SC-10401
    def test_reading_from_root(self):
        self.create_df_from_path_with_exception("/")

    # unit test of SC-10401
    def test_reading_from_empty(self):
        self.create_df_from_path_with_exception("")

    # unit test of SC-10401
    def test_reading_from_untrusted_path(self):
        temp = tempfile.NamedTemporaryFile(delete=True)
        self.create_df_from_path_with_exception(temp.name)
        temp.close()

    # unit test of SC-10401
    def test_reading_from_symbolic_link(self):
        # Create a temp file outside of trusted path
        temp = tempfile.NamedTemporaryFile(delete=True)
        # Create symlink -> temp file
        dir = self.spark._jsparkSession.getTrustedPath()
        link = dir + "/symlink"
        os.symlink(temp.name, link)
        self.create_df_from_path_with_exception(link)
        temp.close()

    # unit test of SC-10401
    def test_reading_by_relative_path_from_trusted_dir(self):
        dir = self.spark._jsparkSession.getTrustedPath()
        self.create_df_from_path_with_exception(dir + "/../")


class FileBasedCollectTests(ReusedPySparkTestCase):
    @classmethod
    def setUpClass(cls):
        ReusedPySparkTestCase.setUpClass()
        cls.spark = SparkSession(cls.sc)
        cls.sc._conf.set("spark.databricks.pyspark.useFileBasedCollect", "true")

    # unit test of SC-8571
    def test_collect_take_and_first(self):
        self.assertEqual(
            self.spark.sql("select * from values 1, 2, 3 as data(a)").collect(),
            [Row(a=1), Row(a=2), Row(a=3)],
        )
        self.assertEqual(
            self.spark.sql("select * from values 1, 2, 3 as data(a)").take(2), [Row(a=1), Row(a=2)]
        )
        self.assertEqual(
            self.spark.sql("select * from values 1, 2, 3 as data(a)").first(), Row(a=1)
        )

    # unit test of SC-18116
    def test_arrow_create_dataframe(self):
        self.assertEqual(
            self.spark.sql("select * from values 1, 2, 3 as data(a)").toPandas().a.tolist(),
            [1, 2, 3],
        )

    # unit test of SC-8571
    def test_toLocalIterator_is_unsupported(self):
        self.assertRaisesRegexp(
            Exception, "file-based collect", lambda: self.spark.range(10).toLocalIterator()
        )

    # unit test of ES-417505
    def test_tail(self):
        self.assertEqual(
            self.spark.sql("select * from values 1, 2, 3 as data(a)").tail(2), [Row(a=2), Row(a=3)]
        )


if __name__ == "__main__":
    from pyspark.databricks.sql.tests.test_edge_changes import *  # noqa: F401,F403

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
