#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2022-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

import unittest

from pyspark.databricks.sql.functions import (
    h3_boundaryasgeojson,
    h3_boundaryaswkb,
    h3_boundaryaswkt,
    h3_centerasgeojson,
    h3_centeraswkb,
    h3_centeraswkt,
    h3_compact,
    h3_distance,
    h3_h3tostring,
    h3_hexring,
    h3_ischildof,
    h3_ispentagon,
    h3_isvalid,
    h3_kring,
    h3_kringdistances,
    h3_longlatash3,
    h3_longlatash3string,
    h3_maxchild,
    h3_minchild,
    h3_pointash3,
    h3_pointash3string,
    h3_polyfillash3,
    h3_polyfillash3string,
    h3_resolution,
    h3_stringtoh3,
    h3_tochildren,
    h3_toparent,
    h3_try_polyfillash3,
    h3_try_polyfillash3string,
    h3_try_validate,
    h3_uncompact,
    h3_validate,
)
from pyspark.sql import Row
from pyspark.sql.functions import hex, unhex
from pyspark.testing.sqlutils import ReusedSQLTestCase


class FunctionsTests(ReusedSQLTestCase):
    def test_h3_h3tostring(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686042433355775,
                    599686042433355776,
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_h3tostring("a"), h3_h3tostring("b")).collect()
        self.assertEqual(results, [Row("85283473fffffff", "852834740000000")])

    def test_h3_stringtoh3(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    "852834740000000",
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_stringtoh3("a"), h3_stringtoh3("b")).collect()
        self.assertEqual(results, [Row(599686042433355775, 599686042433355776)])

    def test_h3_resolution(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_resolution("a"), h3_resolution("b")).collect()
        self.assertEqual(results, [Row(5, 5)])

    def test_h3_isvalid(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_isvalid("a"), h3_isvalid("b")).collect()
        self.assertEqual(results, [Row(True, True)])

        invalid = self.spark.createDataFrame(
            [
                (
                    "852834740000000",
                    "gaea",
                    599686042433355776,
                    -67,
                )
            ],
            ["a", "b", "c", "d"],
        )
        results = invalid.select(
            h3_isvalid("a"), h3_isvalid("b"), h3_isvalid("c"), h3_isvalid("d")
        ).collect()
        self.assertEqual(results, [Row(False, False, False, False)])

    def test_h3_ispentagon(self):
        hexagons = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                )
            ],
            ["a", "b"],
        )
        results = hexagons.select(h3_ispentagon("a"), h3_ispentagon("b")).collect()
        self.assertEqual(results, [Row(False, False)])

        pentagons = self.spark.createDataFrame(
            [
                (
                    "830800fffffffff",
                    590112357393367039,
                )
            ],
            ["a", "b"],
        )
        results = pentagons.select(h3_ispentagon("a"), h3_ispentagon("b")).collect()
        self.assertEqual(results, [Row(True, True)])

    def test_h3_toparent(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                    0,
                )
            ],
            ["a", "b", "c"],
        )
        results = df.select(h3_toparent("a", "c"), h3_toparent("b", "c")).collect()
        self.assertEqual(results, [Row("8029fffffffffff", 577199624117288959)])
        results = df.select(h3_toparent("a", 0), h3_toparent("b", 0)).collect()
        self.assertEqual(results, [Row("8029fffffffffff", 577199624117288959)])

    def test_h3_tochildren(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                    6,
                )
            ],
            ["a", "b", "c"],
        )
        expected1 = [
            "862834707ffffff",
            "86283470fffffff",
            "862834717ffffff",
            "86283471fffffff",
            "862834727ffffff",
            "86283472fffffff",
            "862834737ffffff",
        ]
        expected2 = [
            604189641121202175,
            604189641255419903,
            604189641389637631,
            604189641523855359,
            604189641658073087,
            604189641792290815,
            604189641926508543,
        ]
        results = df.select(h3_tochildren("a", "c"), h3_tochildren("b", "c")).collect()
        self.assertEqual(results, [Row(expected1, expected2)])
        results = df.select(h3_tochildren("a", 6), h3_tochildren("b", 6)).collect()
        self.assertEqual(results, [Row(expected1, expected2)])

    def test_h3_longlatash3(self):
        df = self.spark.createDataFrame(
            [
                (
                    100,
                    45,
                    6,
                ),
            ],
            ["a", "b", "c"],
        )
        results = df.select(h3_longlatash3("a", "b", "c")).collect()
        self.assertEqual(results, [Row(604116085645508607)])
        results = df.select(h3_longlatash3("a", "b", 6)).collect()
        self.assertEqual(results, [Row(604116085645508607)])

    def test_h3_longlatash3string(self):
        df = self.spark.createDataFrame(
            [
                (
                    100,
                    45,
                    6,
                )
            ],
            ["a", "b", "c"],
        )
        results = df.select(h3_longlatash3string("a", "b", "c")).collect()
        self.assertEqual(results, [Row("86240610fffffff")])
        results = df.select(h3_longlatash3string("a", "b", 6)).collect()
        self.assertEqual(results, [Row("86240610fffffff")])

    def test_h3_kring(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686042433355775,
                    "85283473fffffff",
                    1,
                )
            ],
            ["a", "b", "c"],
        )
        expected1 = [
            599686042433355775,
            599686030622195711,
            599686044580839423,
            599686038138388479,
            599686043507097599,
            599686015589810175,
            599686014516068351,
        ]
        expected2 = [
            "85283473fffffff",
            "85283447fffffff",
            "8528347bfffffff",
            "85283463fffffff",
            "85283477fffffff",
            "8528340ffffffff",
            "8528340bfffffff",
        ]
        results = df.select(h3_kring("a", "c"), h3_kring("b", "c")).collect()
        self.assertEqual(results, [Row(expected1, expected2)])
        results = df.select(h3_kring("a", 1), h3_kring("b", 1)).collect()
        self.assertEqual(results, [Row(expected1, expected2)])

    def test_h3_compact(self):
        input1 = [
            599686042433355775,
            599686030622195711,
            599686044580839423,
            599686038138388479,
            599686043507097599,
            599686015589810175,
            599686014516068351,
            599686034917163007,
            599686029548453887,
            599686032769679359,
            599686198125920255,
            599686040285872127,
            599686041359613951,
            599686039212130303,
            599686023106002943,
            599686027400970239,
            599686013442326527,
            599686012368584703,
            599686018811035647,
        ]
        input2 = [
            "85283473fffffff",
            "85283447fffffff",
            "8528347bfffffff",
            "85283463fffffff",
            "85283477fffffff",
            "8528340ffffffff",
            "8528340bfffffff",
            "85283457fffffff",
            "85283443fffffff",
            "8528344ffffffff",
            "852836b7fffffff",
            "8528346bfffffff",
            "8528346ffffffff",
            "85283467fffffff",
            "8528342bfffffff",
            "8528343bfffffff",
            "85283407fffffff",
            "85283403fffffff",
            "8528341bfffffff",
        ]
        df = self.spark.createDataFrame(
            [
                (
                    input1,
                    input2,
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_compact("a"), h3_compact("b")).collect()
        expected1 = [
            599686030622195711,
            599686015589810175,
            599686014516068351,
            599686034917163007,
            599686029548453887,
            599686032769679359,
            599686198125920255,
            599686023106002943,
            599686027400970239,
            599686013442326527,
            599686012368584703,
            599686018811035647,
            595182446027210751,
        ]
        expected2 = [
            "85283447fffffff",
            "8528340ffffffff",
            "8528340bfffffff",
            "85283457fffffff",
            "85283443fffffff",
            "8528344ffffffff",
            "852836b7fffffff",
            "8528342bfffffff",
            "8528343bfffffff",
            "85283407fffffff",
            "85283403fffffff",
            "8528341bfffffff",
            "8428347ffffffff",
        ]
        self.assertEqual(results, [Row(expected1, expected2)])

    def test_h3_uncompact(self):
        input1 = [
            599686030622195711,
            599686015589810175,
            599686014516068351,
            599686034917163007,
            599686029548453887,
            599686032769679359,
            599686198125920255,
            599686023106002943,
            599686027400970239,
            599686013442326527,
            599686012368584703,
            599686018811035647,
            595182446027210751,
        ]
        input2 = [
            "85283447fffffff",
            "8528340ffffffff",
            "8528340bfffffff",
            "85283457fffffff",
            "85283443fffffff",
            "8528344ffffffff",
            "852836b7fffffff",
            "8528342bfffffff",
            "8528343bfffffff",
            "85283407fffffff",
            "85283403fffffff",
            "8528341bfffffff",
            "8428347ffffffff",
        ]
        df = self.spark.createDataFrame(
            [
                (
                    input1,
                    input2,
                    5,
                )
            ],
            ["a", "b", "c"],
        )
        expected1 = [
            599686030622195711,
            599686015589810175,
            599686014516068351,
            599686034917163007,
            599686029548453887,
            599686032769679359,
            599686198125920255,
            599686023106002943,
            599686027400970239,
            599686013442326527,
            599686012368584703,
            599686018811035647,
            599686038138388479,
            599686039212130303,
            599686040285872127,
            599686041359613951,
            599686042433355775,
            599686043507097599,
            599686044580839423,
        ]
        expected2 = [
            "85283447fffffff",
            "8528340ffffffff",
            "8528340bfffffff",
            "85283457fffffff",
            "85283443fffffff",
            "8528344ffffffff",
            "852836b7fffffff",
            "8528342bfffffff",
            "8528343bfffffff",
            "85283407fffffff",
            "85283403fffffff",
            "8528341bfffffff",
            "85283463fffffff",
            "85283467fffffff",
            "8528346bfffffff",
            "8528346ffffffff",
            "85283473fffffff",
            "85283477fffffff",
            "8528347bfffffff",
        ]
        results = df.select(h3_uncompact("a", "c"), h3_uncompact("b", "c")).collect()
        self.assertEqual(results, [Row(expected1, expected2)])
        results = df.select(h3_uncompact("a", 5), h3_uncompact("b", 5)).collect()
        self.assertEqual(results, [Row(expected1, expected2)])

    def test_h3_distance(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686030622195711,
                    "85283447fffffff",
                    599686015589810175,
                    "8528340ffffffff",
                )
            ],
            ["a", "b", "c", "d"],
        )
        results = df.select(
            h3_distance("a", "c"),
            h3_distance("a", "d"),
            h3_distance("b", "c"),
            h3_distance("b", "d"),
        ).collect()
        self.assertEqual(results, [Row(2, 2, 2, 2)])

    def test_h3_validate(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686030622195711,
                    "85283447fffffff",
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_validate("a"), h3_validate("b")).collect()
        self.assertEqual(results, [Row(599686030622195711, "85283447fffffff")])

    def test_h3_try_validate(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686030622195711,
                    "85283447fffffff",
                )
            ],
            ["a", "b"],
        )
        results = df.select(h3_try_validate("a"), h3_try_validate("b")).collect()
        self.assertEqual(results, [Row(599686030622195711, "85283447fffffff")])

        invalid = self.spark.createDataFrame(
            [
                (
                    "852834740000000",
                    "gaea",
                    599686042433355776,
                    -67,
                ),
            ],
            ["a", "b", "c", "d"],
        )
        results = invalid.select(
            h3_try_validate("a"), h3_try_validate("b"), h3_try_validate("c"), h3_try_validate("d")
        ).collect()
        self.assertEqual(results, [Row(None, None, None, None)])

    def test_h3_boundaryas(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686042433355775,
                    "85283473fffffff",
                )
            ],
            ["a", "b"],
        )
        expected_wkt = "POLYGON((-121.915080327 37.271355867,-121.862223289 37.353926451,-121.923549996 37.428341186,-122.037734964 37.420128678,-122.090428929 37.337556084,-122.029101309 37.263197975,-121.915080327 37.271355867))"  # noqa
        expected_geojson = '{"type":"Polygon","coordinates":[[[-121.915080327,37.271355867],[-121.862223289,37.353926451],[-121.923549996,37.428341186],[-122.037734964,37.420128678],[-122.090428929,37.337556084],[-122.029101309,37.263197975],[-121.915080327,37.271355867]]]}'  # noqa
        expected_hexwkb = "01030000000100000007000000646B13AD907A5EC0DE2BFFC9BBA24240B10697AA2E775EC0FF1D42764DAD42409F4271711B7B5EC02EB34CE2D3B64240ED12E93F6A825EC0940FCAC6C6B54240B52A6B96C9855EC044ACAA0935AB4240409BBCCBDC815EC0CC7FA378B0A14240646B13AD907A5EC0DE2BFFC9BBA24240"  # noqa
        results = df.select(h3_boundaryaswkt("a"), h3_boundaryaswkt("b")).collect()
        self.assertEqual(results, [Row(expected_wkt, expected_wkt)])
        results = df.select(h3_boundaryasgeojson("a"), h3_boundaryasgeojson("b")).collect()
        self.assertEqual(results, [Row(expected_geojson, expected_geojson)])
        results = df.select(hex(h3_boundaryaswkb("a")), hex(h3_boundaryaswkb("b"))).collect()
        self.assertEqual(results, [Row(expected_hexwkb, expected_hexwkb)])

    def test_h3_centeras(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686042433355775,
                    "85283473fffffff",
                )
            ],
            ["a", "b"],
        )
        expected_wkt = "POINT(-121.976375973 37.345793375)"
        expected_geojson = '{"type":"Point","coordinates":[-121.976375973,37.345793375]}'
        expected_hexwkb = "0101000000A728A6F17C7E5EC0346612F542AC4240"
        results = df.select(h3_centeraswkt("a"), h3_centeraswkt("b")).collect()
        self.assertEqual(results, [Row(expected_wkt, expected_wkt)])
        results = df.select(h3_centerasgeojson("a"), h3_centerasgeojson("b")).collect()
        self.assertEqual(results, [Row(expected_geojson, expected_geojson)])
        results = df.select(hex(h3_centeraswkb("a")), hex(h3_centeraswkb("b"))).collect()
        self.assertEqual(results, [Row(expected_hexwkb, expected_hexwkb)])

    def test_h3_hexring(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686042433355775,
                    "85283473fffffff",
                    1,
                )
            ],
            ["a", "b", "c"],
        )
        expected1 = [
            599686014516068351,
            599686030622195711,
            599686044580839423,
            599686038138388479,
            599686043507097599,
            599686015589810175,
        ]
        expected2 = [
            "8528340bfffffff",
            "85283447fffffff",
            "8528347bfffffff",
            "85283463fffffff",
            "85283477fffffff",
            "8528340ffffffff",
        ]
        results = df.select(h3_hexring("a", "c"), h3_hexring("b", "c")).collect()
        self.assertEqual(results, [Row(expected1, expected2)])
        results = df.select(h3_hexring("a", 1), h3_hexring("b", 1)).collect()
        self.assertEqual(results, [Row(expected1, expected2)])

    def test_h3_ischildof(self):
        df = self.spark.createDataFrame(
            [
                (
                    608693241318998015,
                    "872834729ffffff",
                    599686042433355775,
                    "85283473fffffff",
                )
            ],
            ["a", "b", "c", "d"],
        )
        results = df.select(
            h3_ischildof("a", "c"),
            h3_ischildof("a", "d"),
            h3_ischildof("b", "c"),
            h3_ischildof("b", "d"),
        ).collect()
        self.assertEqual(results, [Row(True, True, True, True)])
        results = df.select(
            h3_ischildof("c", "a"),
            h3_ischildof("c", "b"),
            h3_ischildof("d", "a"),
            h3_ischildof("d", "b"),
        ).collect()
        self.assertEqual(results, [Row(False, False, False, False)])

    def test_h3_polyfillash3(self):
        input_wkt = "POLYGON((-122.4194 37.7749,-118.2437 34.0522,-74.0060 40.7128,-122.4194 37.7749))"  # noqa
        input_geojson = '{"type":"Polygon","coordinates":[[[-122.4194,37.7749],[-118.2437,34.0522],[-74.0060,40.7128],[-122.4194,37.7749]]]}'  # noqa
        input_hexwkb = "0103000000010000000400000050fc1873d79a5ec0d0d556ec2fe342404182e2c7988f5dc0f46c567dae064140aaf1d24d628052c05e4bc8073d5b444050fc1873d79a5ec0d0d556ec2fe34240"  # noqa
        df = self.spark.createDataFrame(
            [
                (
                    input_wkt,
                    input_geojson,
                    input_hexwkb,
                    2,
                )
            ],
            ["a", "b", "c", "d"],
        )
        expected = [
            586146350232502271,
            586147449744130047,
            586198577034821631,
            586152397546455039,
            586199676546449407,
            586153497058082815,
            586142501941805055,
            586201325813891071,
        ]
        results = df.select(
            h3_polyfillash3("a", "d"), h3_polyfillash3("b", "d"), h3_polyfillash3(unhex("c"), "d")
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])
        results = df.select(
            h3_polyfillash3("a", 2), h3_polyfillash3("b", 2), h3_polyfillash3(unhex("c"), 2)
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])

    def test_h3_polyfillash3string(self):
        input_wkt = "POLYGON((-122.4194 37.7749,-118.2437 34.0522,-74.0060 40.7128,-122.4194 37.7749))"  # noqa
        input_geojson = '{"type":"Polygon","coordinates":[[[-122.4194,37.7749],[-118.2437,34.0522],[-74.0060,40.7128],[-122.4194,37.7749]]]}'  # noqa
        input_hexwkb = "0103000000010000000400000050fc1873d79a5ec0d0d556ec2fe342404182e2c7988f5dc0f46c567dae064140aaf1d24d628052c05e4bc8073d5b444050fc1873d79a5ec0d0d556ec2fe34240"  # noqa
        df = self.spark.createDataFrame(
            [
                (
                    input_wkt,
                    input_geojson,
                    input_hexwkb,
                    2,
                )
            ],
            ["a", "b", "c", "d"],
        )
        expected = [
            "82268ffffffffff",
            "82269ffffffffff",
            "822987fffffffff",
            "8226e7fffffffff",
            "822997fffffffff",
            "8226f7fffffffff",
            "822657fffffffff",
            "8229affffffffff",
        ]
        results = df.select(
            h3_polyfillash3string("a", "d"),
            h3_polyfillash3string("b", "d"),
            h3_polyfillash3string(unhex("c"), "d"),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])
        results = df.select(
            h3_polyfillash3string("a", 2),
            h3_polyfillash3string("b", 2),
            h3_polyfillash3string(unhex("c"), 2),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])

    def test_h3_try_polyfillash3(self):
        input_wkt = "POLYGON((-122.4194 37.7749,-118.2437 34.0522,-74.0060 40.7128,-122.4194 37.7749))"  # noqa
        input_geojson = '{"type":"Polygon","coordinates":[[[-122.4194,37.7749],[-118.2437,34.0522],[-74.0060,40.7128],[-122.4194,37.7749]]]}'  # noqa
        input_hexwkb = "0103000000010000000400000050fc1873d79a5ec0d0d556ec2fe342404182e2c7988f5dc0f46c567dae064140aaf1d24d628052c05e4bc8073d5b444050fc1873d79a5ec0d0d556ec2fe34240"  # noqa
        df = self.spark.createDataFrame(
            [
                (
                    input_wkt,
                    input_geojson,
                    input_hexwkb,
                    2,
                )
            ],
            ["a", "b", "c", "d"],
        )
        expected = [
            586146350232502271,
            586147449744130047,
            586198577034821631,
            586152397546455039,
            586199676546449407,
            586153497058082815,
            586142501941805055,
            586201325813891071,
        ]
        results = df.select(
            h3_try_polyfillash3("a", "d"),
            h3_try_polyfillash3("b", "d"),
            h3_try_polyfillash3(unhex("c"), "d"),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])
        results = df.select(
            h3_try_polyfillash3("a", 2),
            h3_try_polyfillash3("b", 2),
            h3_try_polyfillash3(unhex("c"), 2),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])

        invalid = self.spark.createDataFrame(
            [
                (
                    "I am not a valid WKT or GeoJSON",
                    "020300000000",
                    2,
                )
            ],
            ["a", "b", "c"],
        )
        results = invalid.select(
            h3_try_polyfillash3("a", "c"), h3_try_polyfillash3(unhex("b"), "c")
        ).collect()
        self.assertEqual(results, [Row(None, None)])

    def test_h3_try_polyfillash3string(self):
        input_wkt = "POLYGON((-122.4194 37.7749,-118.2437 34.0522,-74.0060 40.7128,-122.4194 37.7749))"  # noqa
        input_geojson = '{"type":"Polygon","coordinates":[[[-122.4194,37.7749],[-118.2437,34.0522],[-74.0060,40.7128],[-122.4194,37.7749]]]}'  # noqa
        input_hexwkb = "0103000000010000000400000050fc1873d79a5ec0d0d556ec2fe342404182e2c7988f5dc0f46c567dae064140aaf1d24d628052c05e4bc8073d5b444050fc1873d79a5ec0d0d556ec2fe34240"  # noqa
        df = self.spark.createDataFrame(
            [
                (
                    input_wkt,
                    input_geojson,
                    input_hexwkb,
                    2,
                )
            ],
            ["a", "b", "c", "d"],
        )
        expected = [
            "82268ffffffffff",
            "82269ffffffffff",
            "822987fffffffff",
            "8226e7fffffffff",
            "822997fffffffff",
            "8226f7fffffffff",
            "822657fffffffff",
            "8229affffffffff",
        ]
        results = df.select(
            h3_try_polyfillash3string("a", "d"),
            h3_try_polyfillash3string("b", "d"),
            h3_try_polyfillash3string(unhex("c"), "d"),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])
        results = df.select(
            h3_try_polyfillash3string("a", 2),
            h3_try_polyfillash3string("b", 2),
            h3_try_polyfillash3string(unhex("c"), 2),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])

        invalid = self.spark.createDataFrame(
            [
                (
                    "I am not a valid WKT or GeoJSON",
                    "020300000000",
                    2,
                )
            ],
            ["a", "b", "c"],
        )
        results = invalid.select(
            h3_try_polyfillash3string("a", "c"), h3_try_polyfillash3string(unhex("b"), "c")
        ).collect()
        self.assertEqual(results, [Row(None, None)])

    def test_h3_kringdistances(self):
        df = self.spark.createDataFrame(
            [
                (
                    599686042433355775,
                    "85283473fffffff",
                    1,
                )
            ],
            ["a", "b", "c"],
        )
        expected1 = [
            599686042433355775,
            599686030622195711,
            599686044580839423,
            599686038138388479,
            599686043507097599,
            599686015589810175,
            599686014516068351,
        ]
        expected2 = [
            "85283473fffffff",
            "85283447fffffff",
            "8528347bfffffff",
            "85283463fffffff",
            "85283477fffffff",
            "8528340ffffffff",
            "8528340bfffffff",
        ]
        results = df.select(h3_kringdistances("a", "c").alias("p")).select("p.cellid").collect()
        self.assertEqual(results, [Row(expected1)])
        results = df.select(h3_kringdistances("a", "c").alias("p")).select("p.distance").collect()
        self.assertEqual(results, [Row([0, 1, 1, 1, 1, 1, 1])])

        results = df.select(h3_kringdistances("a", 1).alias("p")).select("p.cellid").collect()
        self.assertEqual(results, [Row(expected1)])
        results = df.select(h3_kringdistances("a", 1).alias("p")).select("p.distance").collect()
        self.assertEqual(results, [Row([0, 1, 1, 1, 1, 1, 1])])

        results = df.select(h3_kringdistances("b", "c").alias("p")).select("p.cellid").collect()
        self.assertEqual(results, [Row(expected2)])
        results = df.select(h3_kringdistances("b", "c").alias("p")).select("p.distance").collect()
        self.assertEqual(results, [Row([0, 1, 1, 1, 1, 1, 1])])

        results = df.select(h3_kringdistances("b", 1).alias("p")).select("p.cellid").collect()
        self.assertEqual(results, [Row(expected2)])
        results = df.select(h3_kringdistances("b", 1).alias("p")).select("p.distance").collect()
        self.assertEqual(results, [Row([0, 1, 1, 1, 1, 1, 1])])

    def test_h3_minchild(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                    10,
                )
            ],
            ["a", "b", "c"],
        )
        results = df.select(h3_minchild("a", "c"), h3_minchild("b", "c")).collect()
        self.assertEqual(results, [Row("8a2834700007fff", 622204039496499199)])
        results = df.select(h3_minchild("a", 10), h3_minchild("b", 10)).collect()
        self.assertEqual(results, [Row("8a2834700007fff", 622204039496499199)])

    def test_h3_maxchild(self):
        df = self.spark.createDataFrame(
            [
                (
                    "85283473fffffff",
                    599686042433355775,
                    10,
                )
            ],
            ["a", "b", "c"],
        )
        results = df.select(h3_maxchild("a", "c"), h3_maxchild("b", "c")).collect()
        self.assertEqual(results, [Row("8a2834736db7fff", 622204040416821247)])
        results = df.select(h3_maxchild("a", 10), h3_maxchild("b", 10)).collect()
        self.assertEqual(results, [Row("8a2834736db7fff", 622204040416821247)])

    def test_h3_pointash3(self):
        df = self.spark.createDataFrame(
            [
                (
                    "POINT(-122.4783 37.8199)",
                    '{"type":"Point","coordinates":[-122.4783,37.8199]}',
                    "0101000000516b9a779c9e5ec0c5feb27bf2e84240",
                    13,
                )
            ],
            ["a", "b", "c", "d"],
        )
        expected = 635714569676958015
        results = df.select(
            h3_pointash3("a", "d"), h3_pointash3("b", "d"), h3_pointash3(unhex("c"), "d")
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])
        results = df.select(
            h3_pointash3("a", 13), h3_pointash3("b", 13), h3_pointash3(unhex("c"), 13)
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])

    def test_h3_pointash3string(self):
        df = self.spark.createDataFrame(
            [
                (
                    "POINT(-122.4783 37.8199)",
                    '{"type":"Point","coordinates":[-122.4783,37.8199]}',
                    "0101000000516b9a779c9e5ec0c5feb27bf2e84240",
                    13,
                )
            ],
            ["a", "b", "c", "d"],
        )
        expected = "8d283087022a93f"
        results = df.select(
            h3_pointash3string("a", "d"),
            h3_pointash3string("b", "d"),
            h3_pointash3string(unhex("c"), "d"),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])
        results = df.select(
            h3_pointash3string("a", 13),
            h3_pointash3string("b", 13),
            h3_pointash3string(unhex("c"), 13),
        ).collect()
        self.assertEqual(results, [Row(expected, expected, expected)])


if __name__ == "__main__":
    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
