#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

import unittest

import os

from pyspark import SparkConf, SparkContext
from pyspark.sql import SparkSession
from pyspark.sql.pandas.utils import estimate_pandas_convert_as_arrow_size
from pyspark.sql.tests.pandas.test_pandas_cogrouped_map import BaseCogroupedMapInPandasTests
from pyspark.sql.tests.pandas.test_pandas_grouped_map import BaseGroupedMapInPandasTests

from pyspark.testing.sqlutils import (
    have_pandas,
    have_pyarrow,
    pandas_requirement_message,
    pyarrow_requirement_message,
    SQLTestUtils,
    ReusedSQLTestCase,
)
from pyspark.sql.tests import test_arrow

if have_pandas:
    import pandas as pd
    from pandas.util.testing import assert_frame_equal

if have_pyarrow:
    import pyarrow as pa


@unittest.skipIf(
    not have_pandas or not have_pyarrow, pandas_requirement_message or pyarrow_requirement_message
)
class PandasZeroConfConversionTests(unittest.TestCase):
    """
    Test pandas zero conf conversion.
    """

    @classmethod
    def conf_dict(cls):
        # toPandas() arrow path does not support spark.databricks.pyspark.useFileBasedCollect,
        # disable it here
        return {"spark.databricks.pyspark.useFileBasedCollect": "false"}

    @classmethod
    def setUpClass(cls):
        class ReusedArrowTests(test_arrow.ArrowTests):
            @classmethod
            def conf(inner_cls):
                spark_conf = super(ReusedArrowTests, inner_cls).conf()
                for conf_key, conf_value in cls.conf_dict().items():
                    spark_conf.set(conf_key, conf_value)
                return spark_conf

        ReusedArrowTests.setUpClass()
        cls.arrow_test = ReusedArrowTests()
        cls.spark = cls.arrow_test.spark

        edge_conf_key = "spark.databricks.execution.pandasZeroConfConversion.enabled"
        cls.spark.conf.set(edge_conf_key, "true")
        assert cls.spark.conf.get(edge_conf_key) == "true"
        cls.spark.conf.set(edge_conf_key, "false")
        assert cls.spark.conf.get(edge_conf_key) == "false"

        # enable edge conf
        cls.spark.conf.set(edge_conf_key, "true")
        cls.spark.conf.set(
            "spark.databricks.execution.pandasZeroConfConversion.maxBytesPerSlice", "600"
        )

    @classmethod
    def tearDownClass(cls):
        cls.arrow_test.tearDownClass()

    def test_toPandas_arrow_toggle(self):
        self.arrow_test.test_toPandas_arrow_toggle()

    def test_createDataFrame_empty_partition(self):
        self.arrow_test.test_createDataFrame_empty_partition()

    def test_createDataFrame_gen_correct_partition_num(self):
        default_parallelism = self.spark.sparkContext.defaultParallelism
        max_bytes_per_slice = self.spark._jconf.pandasZeroConfConversionMaxBytesPerSlice()
        pdf = self.arrow_test.create_pandas_data_frame()
        for i in range(6):
            pdf = pdf.append(pdf, ignore_index=True)
        # pdf is large enough to trigger the optimization
        estimated_data_size = estimate_pandas_convert_as_arrow_size(
            pdf, num_sample_blocks=1, sample_ratio_per_block=1.0
        )
        self.assertGreater(estimated_data_size, max_bytes_per_slice * default_parallelism)

        df = self.spark.createDataFrame(pdf, schema=self.arrow_test.schema)

        self.assertEqual(self.arrow_test.schema, df.schema)
        self.assertEqual(df.rdd.getNumPartitions(), default_parallelism)

        pdf_arrow = df.toPandas()
        assert_frame_equal(pdf_arrow, pdf)

    def test_estimate_pandas_df_size(self):
        from distutils.version import LooseVersion

        # test estimate_pandas_size is deterministic by default
        pdf0 = pd.DataFrame({"c1": ["a" * (i % 100 + 1) for i in range(2000)]})
        estimate_pandas_convert_as_arrow_size(pdf0) == estimate_pandas_convert_as_arrow_size(pdf0)

        def assert_estimated_size(pdf, expected_size, with_sample):
            if with_sample:
                est_size = estimate_pandas_convert_as_arrow_size(pdf)
            else:
                est_size = estimate_pandas_convert_as_arrow_size(
                    pdf, num_sample_blocks=1, sample_ratio_per_block=1.0
                )
            self.assertTrue(isinstance(est_size, int))
            self.assertLess(float(abs(est_size - expected_size)) / expected_size, 0.1)

        assert_estimated_size(pd.DataFrame({"c1": [0] * 100}), 800, with_sample=False)
        # pyarrow < 0.16.0 does not have the attribute 'nbytes'
        # size estimation will fallback to pandas memoray_usage path, which is not accurate for
        # short string column bytes
        if LooseVersion(pa.__version__) >= LooseVersion("0.16"):
            # test fix length string column
            # string arrow format bytes includes offset(4 bytes) array and string content array
            # reference: https://arrow.apache.org/docs/format/Columnar.html?\
            # highlight=layout#variable-size-binary-layout
            assert_estimated_size(
                pd.DataFrame({"c1": ["abc"] * 100}), (3 + 4) * 100, with_sample=False
            )
            # test variant length string column
            assert_estimated_size(
                pd.DataFrame({"c1": ["a" * (i + 1) for i in range(200)]}),
                (1 + 200) * 100 + 4 * 200,
                with_sample=False,
            )

        # test fallback to pandas memory_usage path
        # create a pandas dataframe contains a column of object()
        # convert to arrow table will fail and fallback to pandas memory_usage path
        assert_estimated_size(
            pd.DataFrame({"c1": [object()] * 100}), (8 + 16) * 100, with_sample=False
        )

        # Test estimation with sample
        # Test pandas length less than default minimal block size 10
        assert_estimated_size(pd.DataFrame({"c1": [0] * 5}), 40, with_sample=True)
        # less than sample threshold
        assert_estimated_size(pd.DataFrame({"c1": [0] * 100}), 800, with_sample=True)
        # sample ratio > 1/10000
        assert_estimated_size(pd.DataFrame({"c1": [0] * 2000}), 16000, with_sample=True)
        # sample ratio = 1/10000
        assert_estimated_size(pd.DataFrame({"c1": [0] * 200000}), 1600000, with_sample=True)

    def test_create_df_from_empty_pandas_df(self):
        # test for creating spark df from empty pandas dataframe:
        # in case of no schema specified, raise inferring type error.
        # and in case of with schema, it should work.
        from pyspark.sql.types import StructType, StructField, LongType

        df1 = self.spark.range(0)
        pdf1 = df1.toPandas()
        self.assertEqual(len(pdf1), 0)
        with self.assertRaisesRegexp(ValueError, "can not infer schema from empty dataset"):
            self.spark.createDataFrame(pdf1)
        schema1 = StructType([StructField("id", LongType(), True)])
        df1 = self.spark.createDataFrame(pdf1, schema1)
        self.assertEqual(len(df1.toPandas()), 0)

    def test_to_pandas_memoryview_error(self):
        # [ML-13691] value_counts() fails for the data with NULL value dataset generated by
        # spark DataFrame.toPandas()
        result = (
            self.spark.createDataFrame(
                self.spark.sparkContext.parallelize([(i,) for i in range(13)], 1), "id long"
            )
            .selectExpr("IF(id % 3==0, id+1, NULL) AS f1", "(id+1) % 2 AS label")
            .toPandas()["label"]
            .value_counts()
        )
        self.assertEqual(result[0], 6)
        self.assertEqual(result[1], 7)

    def test_zero_size_pdf(self):
        pdf1 = pd.DataFrame({"c1": [None] * 1000})
        df1 = self.spark.createDataFrame(pdf1, schema="c1 int")
        assert df1.rdd.getNumPartitions() == self.spark.sparkContext.defaultParallelism


class PandasZeroConfEncryptionConversionTest(PandasZeroConfConversionTests):
    """
    We need toggle spark.io.encryption.enabled for tests on pandas zero conf conversion
    because encryption on/off have different code path.
    This test covers the encryption path.
    """

    @classmethod
    def conf_dict(cls):
        base_conf = super(PandasZeroConfEncryptionConversionTest, cls).conf_dict()
        custom_conf = {"spark.io.encryption.enabled": "true"}
        return {**base_conf, **custom_conf}


@unittest.skipIf(
    not have_pandas or not have_pyarrow, pandas_requirement_message or pyarrow_requirement_message
)
class PandasZeroConfConversionMaxRpcSizeTests(unittest.TestCase, SQLTestUtils):
    # These tests are separate as 'spark.rpc.message.maxSize' configuration
    # is a static configuration to Spark context.
    # Do not extends ReusedSQLTestCase because it need use local-cluster mode

    # Note that, SPARK-39301 replaced the zero-conf in createDataFrame.
    # We keep the test cases here to make sure no regressions.

    @classmethod
    def setUpClass(cls):
        cls.spark = SparkSession(
            SparkContext(
                "local-cluster[1, 1, 1024]",
                cls.__name__,
                conf=SparkConf()
                .set("spark.rpc.message.maxSize", 1)
                .set("spark.default.parallelism", "1"),
            )
        )

        # Explicitly enable Arrow and disable fallback.
        cls.spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        cls.spark.conf.set("spark.sql.execution.arrow.pyspark.fallback.enabled", "false")

    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, "spark"):
            cls.spark.stop()

    def test_bypass_max_rpc(self):
        self.assertEqual(1, self.spark.sparkContext.defaultParallelism)
        self.assertEqual("1", self.spark.conf.get("spark.rpc.message.maxSize"))
        # construct a pandas df of 16MB size
        pdf = pd.DataFrame({"c1": ["0" * (1024 * 16)] * 1024})
        sdf2 = self.spark.createDataFrame(pdf)
        sdf2.count()


@unittest.skipIf(
    not have_pandas or not have_pyarrow, pandas_requirement_message or pyarrow_requirement_message
)
class PandasZeroConfToPandasMaxResultSizeTests(unittest.TestCase, SQLTestUtils):
    # These tests are separate as 'spark.driver.maxResultSize' and other configurations
    # are static configurations to Spark context.
    # Do not extends ReusedSQLTestCase because it need use local-cluster mode

    @classmethod
    def setUpClass(cls):
        conf = (
            SparkConf()
            .set("spark.default.parallelism", "2")
            .set("spark.driver.maxResultSize", str(cls.max_result_size_in_mb()) + "m")
            .set("spark.databricks.pyspark.useFileBasedCollect", "false")
            .set(
                "spark.databricks.driver.localMaxResultSize",
                str(cls.pandas_max_result_size_in_mb()) + "m",
            )
        )
        # TODO(SC-46097): DBConnect doesn't handle certain spark core conf correctly and will try
        # set them on the server side session conf. We need to disable the check to make the test
        # pass.
        if os.environ.get("PYTEST_DBCONNECT_MODE") is not None:
            conf.set("spark.sql.legacy.setCommandRejectsSparkCoreConfs", "false")
        cls.spark = SparkSession(SparkContext("local-cluster[2, 1, 1024]", cls.__name__, conf=conf))

        # Explicitly enable Arrow and disable fallback.
        cls.spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        cls.spark.conf.set("spark.sql.execution.arrow.pyspark.fallback.enabled", "false")
        cls.spark.conf.set("spark.databricks.execution.pandasZeroConfConversion.enabled", "true")

    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, "spark"):
            cls.spark.stop()

    @classmethod
    def max_result_size_in_mb(cls):
        return 1

    @classmethod
    def pandas_max_result_size_in_mb(cls):
        return 2

    def gen_dataframe(self, size_in_mb):
        return self.spark.range(int(size_in_mb * 1024 * 1024 / 8))

    def test_bypass_max_result_size(self):
        if self.max_result_size_in_mb() > 0:
            result_limit = max(self.max_result_size_in_mb(), self.pandas_max_result_size_in_mb())
        else:
            result_limit = 0  # no limit

        for df_size_in_mb in [0.5, 1.5, 2.5]:
            test_df = self.gen_dataframe(df_size_in_mb)
            if result_limit > 0 and df_size_in_mb > result_limit:
                with self.assertRaises(Exception):
                    test_df.toPandas()
            else:
                test_df.toPandas()  # test no exception raised.


class PandasZeroConfToPandasMaxResultSizeNoLimitTests(PandasZeroConfToPandasMaxResultSizeTests):
    @classmethod
    def max_result_size_in_mb(cls):
        return 0

    @classmethod
    def pandas_max_result_size_in_mb(cls):
        return 2


class PandasZeroConfToPandasMaxResultSizeNoPandasLimitTests(
    PandasZeroConfToPandasMaxResultSizeTests
):
    @classmethod
    def max_result_size_in_mb(cls):
        return 2

    @classmethod
    def pandas_max_result_size_in_mb(cls):
        return 1


@unittest.skipIf(
    not have_pandas or not have_pyarrow, pandas_requirement_message or pyarrow_requirement_message
)
class PandasZeroConfGroupedMapInPandasTests(BaseGroupedMapInPandasTests, ReusedSQLTestCase):
    @classmethod
    def setUpClass(cls):
        super(PandasZeroConfGroupedMapInPandasTests, cls).setUpClass()

        cls.spark.conf.set(
            "spark.databricks.execution.pandasZeroConfConversion.groupbyApply.enabled", "true"
        )
        cls.spark.conf.set(
            "spark.databricks.execution.pandasZeroConfConversion.groupbyApply.maxBytesPerSlice",
            "16",
        )
        # Disable UDF optimization to keep the comparison with regular Python UDFs.
        cls.spark.conf.set("spark.databricks.execution.pythonUDF.arrow.enabled", "false")


@unittest.skipIf(
    not have_pandas or not have_pyarrow, pandas_requirement_message or pyarrow_requirement_message
)
class PandasZeroConfCogroupedMapInPandasTests(BaseCogroupedMapInPandasTests, ReusedSQLTestCase):
    @classmethod
    def setUpClass(cls):
        super(PandasZeroConfCogroupedMapInPandasTests, cls).setUpClass()

        cls.spark.conf.set(
            "spark.databricks.execution.pandasZeroConfConversion.groupbyApply.enabled", "true"
        )
        cls.spark.conf.set(
            "spark.databricks.execution.pandasZeroConfConversion.groupbyApply.maxBytesPerSlice",
            "32",
        )
        # Disable UDF optimization to keep the comparison with regular Python UDFs.
        cls.spark.conf.set("spark.databricks.execution.pythonUDF.arrow.enabled", "false")


if __name__ == "__main__":
    from pyspark.databricks.sql.tests.test_pandas_zero_conf_conversion import *  # noqa: F401,F403

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
