#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#


import os
import sys
import unittest

from pyspark.conf import SparkConf
from pyspark.context import SparkContext
from pyspark.sql import SparkSession
from pyspark.testing.utils import PySparkTestCase, SPARK_HOME, ReusedPySparkTestCase


class AddFileTests(PySparkTestCase):

    # unit test case of [SC-12707] support egg library isolation for python
    def test_sys_path_after_add_library(self):
        isolated_lib1 = "__dbutils_lib_test.egg"
        self.sc.addIsolatedLibraryPath(isolated_lib1)

        lib1 = "userlib-0.1.zip"
        path = os.path.join(SPARK_HOME, "python/test_support", lib1)
        self.sc.addPyFile(path)
        from userlib import UserClass

        self.assertEqual("Hello World from inside a package!", UserClass().hello())

        isolated_lib2 = "__dbutils_lib_test1.egg"
        self.sc.addIsolatedLibraryPath(isolated_lib2)

        lib1_index = self.__getIndexInPathList(sys.path, lib1)
        self.assertNotEqual(lib1_index, -1)
        isolated_lib1_index = self.__getIndexInPathList(sys.path, isolated_lib1)
        self.assertNotEqual(isolated_lib1_index, -1)
        isolated_lib2_index = self.__getIndexInPathList(sys.path, isolated_lib2)
        # verify that in the system path, the cluster wide library occurs after the isolated
        # library, and isolated library added later occurs after the one added before.
        #       isolated_lib2_index < isolated_lib1_index < lib1_index
        self.assertNotEqual(isolated_lib2_index, -1)
        self.assertLess(isolated_lib2_index, isolated_lib1_index)
        self.assertLess(isolated_lib1_index, lib1_index)

        lib1_include_index = self.__getIndexInPathList(self.sc._python_includes, lib1)
        self.assertNotEqual(lib1_include_index, -1)
        isolated_lib1_include_index = self.__getIndexInPathList(
            self.sc._python_includes, isolated_lib1
        )
        self.assertNotEqual(isolated_lib1_include_index, -1)
        isolated_lib2_include_index = self.__getIndexInPathList(
            self.sc._python_includes, isolated_lib2
        )
        self.assertNotEqual(isolated_lib2_include_index, -1)
        # verify that in the python includes, the libraries occurs in the order of added order
        #     isolated_lib1_include_index < lib1_include_index < isolated_lib2_include_index
        self.assertLess(isolated_lib1_include_index, lib1_include_index)
        self.assertLess(isolated_lib1_include_index, isolated_lib2_include_index)

    def __getIndexInPathList(self, pathList, name):
        index = -1
        for i, path in enumerate(pathList):
            (dirname, filename) = os.path.split(path)
            if filename == name:
                index = i
                break
        return index


class ContextTests(unittest.TestCase):

    # unit test case of [SC-9137] Fix Python-side reading of confs. with defaults defined in JVM
    def test_get_configentry(self):
        # Test the use of ConfigEntry
        with SparkContext() as sc:
            # This is a short-term hack until we develop a stable upstream API for this:
            max_task_failures_conf = (
                sc._jvm.__getattr__("org.apache.spark.internal.config.package$")
                .__getattr__("MODULE$")
                .TASK_MAX_FAILURES()
            )
            self.assertEqual(sc._conf.get(max_task_failures_conf), 4)
            # Passing a default with a ConfigEntry is forbidden:
            self.assertRaises(Exception, lambda: sc._conf.get(max_task_failures_conf, 10))


# WARNING(greg) 2018-09-24: These tests set Spark confs that can interfere with other tests because
# of the bug described in SC-11328. Until that bug is fixed, we must run these tests last.
# unit tests of [SC-12408] Implement safe Spark file APIs for credential passthrough (RDD API)
class ZZZ_ContextLocalFileSystemTests(unittest.TestCase):
    def get_spark_session(self):
        conf = (
            SparkConf()
            .set(
                # Disable the local filesystem
                "spark.databricks.pyspark.onlyAllowTrustedFilesystems",
                "true",
            )
            .set(
                # Use trusted data transfer to make sc.parallelize work
                "spark.databricks.pyspark.trustedJVMDataTransfer.enabled",
                "true",
            )
            .set(
                # Enforce that the local filesystem is used by default when no scheme is specified
                "spark.hadoop.fs.default.name",
                "file:///",
            )
        )
        sc = SparkContext(conf=conf)
        # We need to initialize the SparkSession
        return SparkSession(sc)

    # method should look like this:
    # def method(sc: SparkContext, path: String): Unit = {
    #   sc.callSomeMethod(path, otherArgs)
    # }
    def local_file_method_test(self, method):
        with self.get_spark_session() as spark:
            base_path = "/local/path"
            self.assertRaisesRegexp(
                Exception,
                ".*%s.*untrusted filesystem.*" % base_path,
                lambda: method(spark._sc, base_path),
            )
            self.assertRaisesRegexp(
                Exception,
                ".*%s.*untrusted filesystem.*" % base_path,
                lambda: method(spark._sc, "file://" + base_path),
            )

    def test_pickle_file(self):
        self.local_file_method_test(lambda sc, path: sc.pickleFile(path))

    def test_text_file(self):
        self.local_file_method_test(lambda sc, path: sc.textFile(path))

    def test_whole_text_files(self):
        self.local_file_method_test(lambda sc, path: sc.wholeTextFiles(path))

    def test_binary_files(self):
        self.local_file_method_test(lambda sc, path: sc.binaryFiles(path))

    def test_binary_records(self):
        self.local_file_method_test(lambda sc, path: sc.binaryRecords(path, 1))

    def test_sequence_file(self):
        self.local_file_method_test(lambda sc, path: sc.sequenceFile(path))

    def test_new_api_hadoop_file(self):
        self.local_file_method_test(
            lambda sc, path: sc.newAPIHadoopFile(
                path,
                "org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat",
                "org.apache.hadoop.io.IntWritable",
                "org.apache.hadoop.io.Text",
            )
        )

    def test_hadoop_file(self):
        self.local_file_method_test(
            lambda sc, path: sc.hadoopFile(
                path,
                "org.apache.hadoop.mapred.SequenceFileInputFormat",
                "org.apache.hadoop.io.IntWritable",
                "org.apache.hadoop.io.Text",
            )
        )

    def test_checkpoint_file(self):
        def method(sc, path):
            parCollection = sc.parallelize([1, 2, 3, 4])
            sc._checkpointFile(path, parCollection._jrdd_deserializer)

        self.local_file_method_test(method)


# WARNING(greg) 2018-09-24: These tests set Spark confs that can interfere with other tests because
# of the bug described in SC-11328. Until that bug is fixed, we must run these tests last.
# unit tests of [SC-12408] Implement safe Spark file APIs for credential passthrough (RDD API)
class ZZZ_RDDLocalFileSystemTests(unittest.TestCase):
    def get_spark_session(self):
        conf = (
            SparkConf()
            .set(
                # Disable the local filesystem
                "spark.databricks.pyspark.onlyAllowTrustedFilesystems",
                "true",
            )
            .set(
                # Use trusted data transfer to make sc.parallelize work
                "spark.databricks.pyspark.trustedJVMDataTransfer.enabled",
                "true",
            )
            .set(
                # Enforce that the local filesystem is used by default when no scheme is specified
                "spark.hadoop.fs.default.name",
                "file:///",
            )
        )
        sc = SparkContext(conf=conf)
        # We need to initialize the SparkSession
        return SparkSession(sc)

    # method should look like this:
    # def method(sc: SparkContext, path: String): Unit = {
    #   sc.callSomeMethod(path, otherArgs)
    # }
    def local_file_method_test(self, method):
        with self.get_spark_session() as spark:
            base_path = "/local/path"
            self.assertRaisesRegexp(
                Exception,
                ".*%s.*untrusted filesystem.*" % base_path,
                lambda: method(spark._sc, base_path),
            )
            self.assertRaisesRegexp(
                Exception,
                ".*%s.*untrusted filesystem.*" % base_path,
                lambda: method(spark._sc, "file://" + base_path),
            )

    def test_save_as_new_api_hadoop_file(self):
        def method(sc, path):
            data = [(1, ""), (1, "a"), (2, "bcdf")]
            sc.parallelize(data).saveAsNewAPIHadoopFile(
                path,
                "org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat",
                "org.apache.hadoop.io.IntWritable",
                "org.apache.hadoop.io.Text",
            )

        self.local_file_method_test(method)

    def test_save_as_hadoop_file(self):
        def method(sc, path):
            dict_data = [(1, {}), (1, {"row1": 1.0}), (2, {"row2": 2.0})]
            sc.parallelize(dict_data).saveAsHadoopFile(
                path,
                "org.apache.hadoop.mapred.SequenceFileOutputFormat",
                "org.apache.hadoop.io.IntWritable",
                "org.apache.hadoop.io.MapWritable",
            )

        self.local_file_method_test(method)

    def test_save_as_sequence_file(self):
        def method(sc, path):
            ei = [(1, "aa"), (1, "aa"), (2, "aa"), (2, "bb"), (2, "bb"), (3, "cc")]
            sc.parallelize(ei).saveAsSequenceFile(path)

        self.local_file_method_test(method)

    def test_save_as_pickle_file(self):
        self.local_file_method_test(
            lambda sc, path: sc.parallelize(range(10)).saveAsPickleFile(path)
        )

    def test_save_as_text_file(self):
        self.local_file_method_test(lambda sc, path: sc.parallelize(range(10)).saveAsTextFile(path))


# WARNING(greg) 2018-09-24: These tests set Spark confs that can interfere with other tests because
# of the bug described in SC-11328. Until that bug is fixed, we must run these tests last.
# unit tests of [SC-12049] Use file-based collect for RDD API when credential passthrough enabled
class ZZZ_FileBasedCollectTests(unittest.TestCase):
    def get_spark_session(self):
        conf = (
            SparkConf()
            .set(
                # Disable the local filesystem
                "spark.databricks.pyspark.onlyAllowTrustedFilesystems",
                "true",
            )
            .set(
                # Use trusted data transfer to make sc.parallelize work
                "spark.databricks.pyspark.trustedJVMDataTransfer.enabled",
                "true",
            )
            .set(
                # Use file-based collect
                "spark.databricks.pyspark.useFileBasedCollect",
                "true",
            )
            .set(
                # Enforce that the local filesystem is used by default when no scheme is specified
                "spark.hadoop.fs.default.name",
                "file:///",
            )
        )
        sc = SparkContext(conf=conf)
        # We need to initialize the SparkSession
        return SparkSession(sc)

    def collect_data_test(self, data):
        with self.get_spark_session() as spark:
            collected = spark._sc.parallelize(data).collect()
            self.assertEqual(data, collected)

    def take_data_test(self, data, num, expected):
        with self.get_spark_session() as spark:
            actual = spark._sc.parallelize(data).take(num)
            self.assertEqual(expected, actual)

    def test_simple_list(self):
        self.collect_data_test([1, 2, 3, 4])

    def test_range(self):
        self.collect_data_test(list(range(10)))

    def test_object_list(self):
        self.collect_data_test([(1, {}), (1, {"row1": 1.0}), (2, {"row2": 2.0})])

    def test_take_list(self):
        self.take_data_test([1, 2, 3, 4], 0, [])
        self.take_data_test([1, 2, 3, 4], 1, [1])
        self.take_data_test([1, 2, 3, 4], 2, [1, 2])
        self.take_data_test([1, 2, 3, 4], 10, [1, 2, 3, 4])

    def test_take_range(self):
        self.take_data_test(list(range(10)), 4, [0, 1, 2, 3])

    def test_take_object_list(self):
        self.take_data_test([("a", 1), ("b", 2), ("c", 3), ("d", 4)], 2, [("a", 1), ("b", 2)])

    def test_take_large_list(self):
        self.take_data_test(list(range(1000)), 10, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])


@unittest.skipIf(
    os.environ.get("PYTEST_DBCONNECT_MODE") is not None, "dbconnect handles job groups differently"
)
class RDDCollectWithJobGroupTest(ReusedPySparkTestCase):
    def check_collect_with_job_group(self):
        import time
        import threading

        group_A_name = "group_A"
        group_B_name = "group_B"

        def map_func(x):
            time.sleep(3)
            return x + 1

        num_threads = 4
        thread_list = []
        # an array which record whether job is cancelled.
        # the index of the array is the thread index which job run in.
        is_job_cancelled = [False for x in range(num_threads)]

        def run_job(job_group, index):
            try:
                result = (
                    self.sc.parallelize([3])
                    .map(map_func)
                    .collectWithJobGroup(job_group, "test rdd collect with setting job group")
                )
                is_job_cancelled[index] = False
                return result
            except Exception:
                is_job_cancelled[index] = True
                return None

        def launch_job_thread(job_group, index):
            thread = threading.Thread(target=run_job, args=(job_group, index))
            thread.start()
            return thread

        # test job succeeded when not cancelled.
        run_job(group_A_name, 0)
        self.assertFalse(is_job_cancelled[0], "job didn't succeeded.")

        # launch spark job in multiple threads and cancel half of them.
        for i in range(num_threads):
            if i % 2 == 0:
                thread = launch_job_thread(group_A_name, i)
            else:
                thread = launch_job_thread(group_B_name, i)
            thread_list.append(thread)

        time.sleep(1)
        self.sc.cancelJobGroup(group_A_name)

        for i in range(num_threads):
            thread_list[i].join()
            if i % 2 == 0:
                # make sure group A job being cancelled.
                self.assertTrue(is_job_cancelled[i], "Job in group A wasn't cancelled.")
            else:
                # make sure group B job succeeded.
                self.assertFalse(is_job_cancelled[i], "Job in group B didn't succeeded.")

    def test_collect_with_job_group(self):
        self.check_collect_with_job_group()

    def test_file_based_collect_with_job_group(self):
        self.sc._conf.set("spark.databricks.pyspark.useFileBasedCollect", "true")
        self.check_collect_with_job_group()


if __name__ == "__main__":
    from pyspark.databricks.tests.test_edge_changes import *  # noqa: F401,F403

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
