#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

from pyspark.databricks.testing.utils import UsageLoggingTestCase
from pyspark.sql.window import Window, WindowSpec


class PysparkUsageLoggingTest(UsageLoggingTestCase):
    """
    Test suite for pyspark usage logging.
    """

    def test_pyspark_usage_logging(self):
        from pyspark.instrumentation_utils import _auto_patch_spark
        from pyspark.sql import _class_allow_list, _class_property_allow_list

        _auto_patch_spark([], _class_allow_list, _class_property_allow_list)

        from pyspark.databricks import usage_logger

        usage_logger.get_logger().enabled = True

        def function_with_logging():
            df = self.spark.createDataFrame([(1, 2)], ["a", "b"])

            # df.columns won't be logged since it is not in the class property allow list.
            self.assertEquals(list(df.columns), ["a", "b"])

            self.assertFalse(df.rdd.map(lambda x: 2 * x).isLocallyCheckpointed())

            self.assertRaises(AttributeError, lambda: df.c)

            # SC-98102: The usage logger attachment logic should handle static methods properly
            self.assertTrue(isinstance(Window.partitionBy("id"), WindowSpec))
            self.assertTrue(isinstance(Window().partitionBy("id"), WindowSpec))

        usage_records = list(self.track_usage(function_with_logging))

        self.assertEquals(
            usage_records[0].metric(),
            self.metric_definitions().EVENT_INITIAL_CONFIG_LOG().name(),
        )

        self.assertTrue(
            all(
                record.metric() == self.metric_definitions().EVENT_PYSPARK_FUNCTION_CALLED().name()
                for record in usage_records[1:]
            )
        )

        success_records = [
            record
            for record in usage_records[1:]
            if self.get_tag_value(record, self.tag_definitions().TAG_STATUS()) == "success"
        ]
        self.assertEquals(len(success_records), 6)

        self.assertTrue(
            self.get_tag_value(
                success_records[0], self.tag_definitions().TAG_PYSPARK_FUNCTION()
            ).startswith("createDataFrame")
        )
        self.assert_tag(
            success_records[1],
            self.tag_definitions().TAG_PYSPARK_PROPERTY(),
            "rdd",
        )
        self.assertTrue(
            self.get_tag_value(
                success_records[2], self.tag_definitions().TAG_PYSPARK_FUNCTION()
            ).startswith("map")
        )
        self.assertTrue(
            self.get_tag_value(
                success_records[3], self.tag_definitions().TAG_PYSPARK_FUNCTION()
            ).startswith("isLocallyCheckpointed")
        )
        self.assertTrue(
            self.get_tag_value(
                success_records[4], self.tag_definitions().TAG_PYSPARK_FUNCTION()
            ).startswith("partitionBy")
        )
        self.assertTrue(
            self.get_tag_value(
                success_records[5], self.tag_definitions().TAG_PYSPARK_FUNCTION()
            ).startswith("partitionBy")
        )

        failure_records = [
            record
            for record in usage_records[1:]
            if self.get_tag_value(record, self.tag_definitions().TAG_STATUS()) == "failure"
        ]
        self.assertEquals(len(failure_records), 1)
        self.assertTrue(
            self.get_tag_value(
                failure_records[0], self.tag_definitions().TAG_PYSPARK_FUNCTION()
            ).startswith("__getattr__")
        )


if __name__ == "__main__":
    import unittest
    from pyspark.databricks.tests.test_usage_logging import *  # noqa: F401,F403

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
