#
# DATABRICKS CONFIDENTIAL & PROPRIETARY
# __________________
#
# Copyright 2020-present Databricks, Inc.
# All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property of Databricks, Inc.
# and its suppliers, if any.  The intellectual and technical concepts contained herein are
# proprietary to Databricks, Inc. and its suppliers and may be covered by U.S. and foreign Patents,
# patents in process, and are protected by trade secret and/or copyright law. Dissemination, use,
# or reproduction of this information is strictly forbidden unless prior written permission is
# obtained from Databricks, Inc.
#
# If you view or obtain a copy of this information and believe Databricks, Inc. may not have
# intended it to be made available, please promptly report it to Databricks Legal Department
# @ legal@databricks.com.
#

import inspect
import time
import unittest

from pyspark.databricks.testing.utils import UsageLoggingTestCase
from pyspark.databricks.utils.instrumentation import instrumented


class PythonUsageLoggingTest(UsageLoggingTestCase):
    """
    Test suite for Python wrapper for Databricks usage logging.
    """

    def test_basic_usage_logging(self):
        logger = self._usage_logging

        def function_with_logging():
            logger.recordProductEvent(
                logger.metricDefinitions().EVENT_ML_ALGORITHM(),
                {logger.tagDefinitions().TAG_ML_RUN_UUID(): "fakeTagEntry"},
                "my big blob",
            )

        usage_records = self.track_usage(function_with_logging)
        self.assertEqual(
            len(usage_records),
            1,
            "Expected 1 usage record but found {n}.".format(n=len(usage_records)),
        )
        self.assert_tag(usage_records[0], logger.tagDefinitions().TAG_ML_RUN_UUID(), "fakeTagEntry")


class InstrumentationTest(UsageLoggingTestCase):
    """
    Test suite for decorator `pyspark.databricks.utils.instrumentation.instrumented`.
    """

    def setUp(self):
        super(InstrumentationTest, self).setUp()
        self.test_err2_info = "InstrumentTestClassFn2TestError"
        self.test_err3_info = "InstrumentTestClassFn3TestError"

        class InstrumentTestClass:
            @instrumented
            def fn1(self, a, b, c, d):
                """
                Doc of func1
                """
                time.sleep(3)
                return a + b + c + d

            @instrumented
            def fn2(_):
                raise RuntimeError(self.test_err2_info)

            @instrumented
            def fn3(_):
                test_err3_info = self.test_err3_info

                def mapper(_):
                    raise RuntimeError(test_err3_info)

                self.sc.parallelize(range(10)).map(mapper).collect()

        self.instrument_test = InstrumentTestClass()

    def test_run_successfully(self):
        result = []
        usage_records = self.track_usage(
            lambda: result.append(self.instrument_test.fn1(1, 20, c=300, d=4000))
        )
        self.assertEqual(result[0], 4321, "call InstrumentTestClass.fn1 get wrong result.")
        self.assertEqual(
            len(usage_records),
            1,
            "Expected 1 usage record but found {n}.".format(n=len(usage_records)),
        )
        usage_record = usage_records[0]
        self.assert_tag(
            usage_record, self.tag_definitions().TAG_CLASS_NAME(), "InstrumentTestClass"
        )
        self.assert_tag(usage_record, self.tag_definitions().TAG_FUNCTION_NAME(), "fn1")
        self.assert_tag(usage_record, self.tag_definitions().TAG_STATUS(), "finished")
        self.assertEqual(self.get_blob(usage_record), "")
        self.assert_quantity(
            usage_record,
            self.metric_definitions().METRIC_FUNCTION_DURATION(),
            expected_quantity=3000,
            quantity_abs_tol=500,
        )

    def test_run_failed(self):
        usage_records = self.track_usage(lambda: self.instrument_test.fn2())
        self.assertEqual(
            len(usage_records),
            1,
            "Expected 1 usage record but found {n}.".format(n=len(usage_records)),
        )
        usage_record = usage_records[0]
        self.assert_tag(
            usage_record, self.tag_definitions().TAG_CLASS_NAME(), "InstrumentTestClass"
        )
        self.assert_tag(usage_record, self.tag_definitions().TAG_FUNCTION_NAME(), "fn2")
        self.assert_tag(usage_record, self.tag_definitions().TAG_STATUS(), "failed")
        self.assert_tag(usage_record, self.tag_definitions().TAG_ERROR(), "RuntimeError")
        err_msg = self.get_blob(usage_record)
        self.assertTrue("raise RuntimeError" in err_msg, "Error message should should be in log")
        # Exception message should be in log
        self.assertTrue(self.test_err2_info in err_msg, "Exception message should be in log")

    def test_run_fail_with_py4jJavaError(self):
        usage_records = self.track_usage(lambda: self.instrument_test.fn3())

        for usage_record in usage_records:
            if usage_record.metric() == self.metric_definitions().METRIC_FUNCTION_DURATION():
                self.assert_tag(usage_record, self.tag_definitions().TAG_STATUS(), "failed")
                self.assert_tag(usage_record, self.tag_definitions().TAG_ERROR(), "Py4JJavaError")
                self.assert_tag(usage_record, self.tag_definitions().TAG_FUNCTION_NAME(), "fn3")
                err_msg = self.get_blob(usage_record)
                self.assertTrue(
                    "Py4JJavaError: An error occurred while calling" in err_msg,
                    "Error message should should be in log",
                )
                # Py4JJavaError jvm side exception message should be in log
                self.assertTrue(
                    self.test_err3_info in err_msg, "Exception message should be in log"
                )

    def test_signature_and_doc(self):
        fn1 = self.instrument_test.fn1
        self.assertEqual(fn1.__doc__.strip(), "Doc of func1")
        sig = inspect.signature(fn1)
        self.assertEqual(list(sig.parameters.keys()), ["a", "b", "c", "d"])

    def test_truncate_error_message(self):
        def test_fn():
            self._usage_logging.recordFunctionDuration(
                1000.0,
                {
                    self._usage_logging.tagDefinitions().TAG_CLASS_NAME(): "classA",
                    self._usage_logging.tagDefinitions().TAG_FUNCTION_NAME(): "testfun",
                    self._usage_logging.tagDefinitions().TAG_STATUS(): "failed",
                    self._usage_logging.tagDefinitions().TAG_ERROR(): "RuntimeError",
                },
                "a" * 5000,
            )

        usage_records = self.track_usage(lambda: test_fn())
        self.assertEqual(
            len(usage_records),
            1,
            "Expected 1 usage record but found {n}.".format(n=len(usage_records)),
        )
        err_msg = self.get_blob(usage_records[0])
        assert len(err_msg) <= 20000 + len("[TRUNCATED]")


if __name__ == "__main__":
    from pyspark.databricks.tests.test_usage_logging_utils import *  # noqa: F401,F403

    try:
        import xmlrunner

        unittest.main(
            testRunner=xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2),
            verbosity=2,
        )
    except ImportError:
        unittest.main(verbosity=2)
