#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import stat
import tempfile
import unittest
from pyspark import SparkConf, SparkContext, TaskContext
from pyspark.testing.utils import SPARK_HOME


class TaskContextTestsWithMultipleResources(unittest.TestCase):
    def setUp(self):
        class_name = self.__class__.__name__
        self.tempFile = tempfile.NamedTemporaryFile(delete=False)
        self.tempFile.write(b'echo {\\"name\\": \\"gpu\\", \\"addresses\\": [\\"2\\", \\"3\\"]}')
        self.tempFile.close()
        # create temporary directory for Worker resources coordination
        self.tempdir = tempfile.NamedTemporaryFile(delete=False)
        os.unlink(self.tempdir.name)
        os.chmod(
            self.tempFile.name,
            stat.S_IRWXU | stat.S_IXGRP | stat.S_IRGRP | stat.S_IROTH | stat.S_IXOTH,
        )
        conf = SparkConf().set("spark.test.home", SPARK_HOME)
        conf = conf.set("spark.worker.resource.gpu.discoveryScript", self.tempFile.name)
        conf = conf.set("spark.worker.resource.gpu.amount", 2)
        conf = conf.set("spark.task.resource.gpu.amount", "2")
        conf = conf.set("spark.executor.resource.gpu.amount", "2")
        # local cluster with 1 worker node, and 1 cores per worker node
        self.sc = SparkContext("local-cluster[1,1,1024]", class_name, conf=conf)

    def test_resources(self):
        """Test the resources are available."""
        rdd = self.sc.parallelize(range(12))
        resources = rdd.map(lambda x: TaskContext.get().resources()).take(1)[0]
        self.assertEqual(len(resources), 1)
        self.assertTrue("gpu" in resources)
        self.assertEqual(resources["gpu"].name, "gpu")
        self.assertEqual(resources["gpu"].addresses, ["0", "1"])

    def test_cuda_visible_devices(self):
        """Test that CUDA environment variables are set correctly"""
        rdd = self.sc.parallelize(range(1))

        def get_cuda_visible_devices():
            return os.environ["CUDA_VISIBLE_DEVICES"]

        cuda_visible_devices = rdd.map(lambda _: get_cuda_visible_devices()).collect()[0]
        addresses = set(cuda_visible_devices.split(","))
        self.assertEqual(len(addresses), 2)
        self.assertEqual(addresses, set(["2", "3"]))

        def get_cuda_device_order():
            return os.environ["CUDA_DEVICE_ORDER"]

        cuda_device_orders = rdd.map(lambda _: get_cuda_device_order()).collect()
        for cuda_device_order in cuda_device_orders:
            self.assertEqual(cuda_device_order, "PCI_BUS_ID")

    def tearDown(self):
        os.unlink(self.tempFile.name)
        self.sc.stop()


if __name__ == "__main__":
    import unittest
    from pyspark.tests.test_taskcontext_multigpu import *  # noqa: F401

    try:
        import xmlrunner

        testRunner = xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2)
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)
