import json
import os
import sys
import threading
import unittest
from http.server import HTTPServer, BaseHTTPRequestHandler

from pyspark import SparkConf, SparkContext


# Tracking the state of the test server for ExecutorSendingCredentialsToWsfsTest
class WsfsTestServerState:
    def __init__(self):
        self.requests = []


# A test server for ExecutorSendingCredentialsToWsfsTest
class WsfsTestServerHandler(BaseHTTPRequestHandler):
    def __init__(self, server_state, *args):
        self.server_state = server_state
        super().__init__(*args)

    def _send_response(self, code):
        self.send_response(code)
        self.end_headers()

    def do_POST(self):
        if self.path != "/api/1/init":
            self._send_response(404)
            return
        length = int(self.headers.get("content-length"))
        json.loads(self.rfile.read(length))

        self._send_response(200)

    def do_PUT(self):
        supported_path = "/api/1/pid/"
        if supported_path not in self.path or len(self.path) == len(supported_path):
            self._send_response(404)
            return

        pid = self.path[self.path.rfind("/") + 1 :]
        length = int(self.headers.get("content-length"))
        body = json.loads(self.rfile.read(length))
        credential = body["apiToken"]
        start_time = body["procStartTime"]
        command_origin = body["commandOrigin"]
        command_id = body["commandId"]
        self.server_state.requests.append((pid, credential, start_time, command_origin, command_id))
        self._send_response(200)


class ExecutorSendingCredentialsToWsfsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.server_state = WsfsTestServerState()

        def handler(*args):
            WsfsTestServerHandler(cls.server_state, *args)

        cls.server = HTTPServer(("localhost", 0), handler)  # pick a random port

        cls.t = threading.Thread(target=cls.server.serve_forever)
        cls.t.daemon = True
        cls.t.start()

    def setUp(self):
        # reset state for each test
        self.__class__.server_state = WsfsTestServerState()

        # Prevent tests from using /Workspace prefix, as we may
        # not have permissions in Jenkins to mkdir under /
        os.environ["WSFS_TEST_DIR"] = "/tmp"

        class_name = self.__class__.__name__
        conf = (
            SparkConf()
            .set("spark.databricks.enableWsfs", "true")
            .set("spark.databricks.wsfsRpcPort", self.server.server_port)
        )
        self.sc = SparkContext("local[1]", class_name, conf=conf)
        self.sc._jvm.org.apache.spark.TestUtils.putCredentialForTesting(
            "spark.databricks.api.token", "test_credential"
        )
        self.sc.setLocalProperty("spark.databricks.api.url", "http://localhost")
        self.sc.setLocalProperty("spark.databricks.notebook.path", "/Repos/user@domain/foo/bar.nb")
        self.sc.setLocalProperty("spark.jobGroup.id", "1_1_1")

        # Setup temporary directories that it should "chdir" into.
        os.makedirs("/tmp/Repos/user@domain/foo", exist_ok=True)
        os.makedirs("/tmp/Repos/user@domain/foo2", exist_ok=True)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires Linux")
    def test_pid_post_request(self):
        pids = self.sc.parallelize(range(0, 1)).map(lambda x: os.getpid()).collect()
        pid = str(pids[0])
        start_time = get_proc_start_time(pid)
        # The wsfs http client should set the credential before the executor executes
        # a command and reset the credential after the command is finished
        self.assertEqual(
            self.server_state.requests,
            [
                (pid, "test_credential", start_time, "PythonWorker", "1_1_1"),
                (pid, "", 0, "Unknown", ""),
            ],
        )

    def test_reuse_worker_of_parallelize_range(self):
        rdd = self.sc.parallelize(range(20), 8)
        previous_pids = rdd.map(lambda x: os.getpid()).collect()
        # should not re-use workers if notebook dir is different
        self.sc.setLocalProperty("spark.databricks.notebook.path", "/Repos/user@domain/foo2/bar.nb")
        current_pids = rdd.map(lambda x: os.getpid()).collect()
        for pid in current_pids:
            self.assertNotIn(pid, previous_pids)

    def tearDown(self):
        self.sc._jvm.org.apache.spark.TestUtils.removeCredentialForTesting(
            "spark.databricks.api.token"
        )
        self.sc.stop()

    @classmethod
    def tearDownClass(cls):
        cls.server.shutdown()
        cls.t.join()


# Implementation based on psutils
# https://github.com/giampaolo/psutil/blob/master/psutil/_pslinux.py#L1607
def get_proc_start_time(pid):
    with open("/proc/%s/stat" % pid) as proc:
        data = proc.read()
        rpar = data.rfind(")")
        fields = data[rpar + 2 :].split()
        return int(fields[19])


if __name__ == "__main__":
    import unittest
    from pyspark.tests.test_wsfs import *  # noqa: F401

    try:
        import xmlrunner  # type: ignore[import]

        testRunner = xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2)
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)
