import unittest
import os
import importlib.abc
from unittest.mock import patch, MagicMock

from pyspark.wsfs_import_hook import WsfsImportHook


class WsfsImportHookSuite(unittest.TestCase):
    def setUp(self):
        super(WsfsImportHookSuite, self).setUp()

        self.import_path = "/Workspace/Users/user@databricks.com"
        os.makedirs(self.import_path, exist_ok=True)
        os.chdir(self.import_path)

        with open(self.import_path + "/wsfs_lib.py", "w+") as f:
            f.write("def foo(): return 'foo' ")

        self.wsfs_import_hook = WsfsImportHook(self.import_path)

    def test_is_path_entry_finder(self):
        self.assertTrue(isinstance(self.wsfs_import_hook, importlib.abc.PathEntryFinder))  # noqa

    def test_only_support_files_in_workspace(self):
        self.assertRaises(ImportError, lambda: WsfsImportHook("/path/to/random/file"))  # noqa
        self.assertRaises(
            ImportError, lambda: WsfsImportHook("/Workspace/Repos/path/to/random/file")  # noqa
        )

    def test_handle_notebook_origin(self):
        self.wsfs_import_hook.get_filename = MagicMock(
            return_value="IPython/core/interactiveshell.py"
        )
        self.assertIsNotNone(self.wsfs_import_hook.find_spec("wsfs_lib", None))

    def test_handle_site_package_origin(self):
        self.wsfs_import_hook = WsfsImportHook(
            # noqa
            path=self.import_path,
            site_packages=["site-packages-mock"],
        )
        self.wsfs_import_hook.get_filename = MagicMock(
            return_value="site-packages-mock/file.py"
        )  # noqa
        self.assertIsNone(self.wsfs_import_hook.find_spec("wsfs_lib", None))

        self.wsfs_import_hook.get_filename.assert_called_once

    def test_handle_wsfs_file_origin(self):
        self.wsfs_import_hook.get_filename = MagicMock(return_value="random-file.py")  # noqa
        self.assertIsNotNone(self.wsfs_import_hook.find_spec("wsfs_lib", None))

        self.wsfs_import_hook.get_filename.assert_called_once

    def test_file_finder_not_found(self):
        with patch("sys.path_hooks") as path_hooks_patch:
            path_hooks_patch.return_value = []
            self.assertRaises(ImportError, lambda: WsfsImportHook(self.import_path))  # noqa

    def test_enforces_max_recursion_limit(self):
        import_hook = WsfsImportHook(path=self.import_path, max_recursion_depth=1)  # noqa
        import_hook.get_filename = MagicMock(return_value="unknown-package.py")  # noqa

        self.assertIsNotNone(import_hook.find_spec("wsfs_lib", None))

        # We should have only invoked get_filename once, since we would
        # have exited the loop early because of the max_recursion_depth
        import_hook.get_filename.assert_called_once

    def test_fail_open_find_spec(self):
        with patch.object(
            WsfsImportHook, "_WsfsImportHook__is_user_import"
        ) as user_import_mock:  # noqa
            user_import_mock.side_effect = Exception("Fail")
            self.assertIsNone(self.wsfs_import_hook.find_spec("wsfs_lib", None))  # noqa


if __name__ == "__main__":
    unittest.main()
