import os
import tempfile
import threading
import unittest
from http.server import HTTPServer, BaseHTTPRequestHandler

from pyspark import SparkConf, SparkContext
from pyspark.wsfs_import_hook import WsfsImportHook


# A dummy wsfs server
class WsfsDummyServerHandler(BaseHTTPRequestHandler):
    def _send_response(self, code):
        self.send_response(code)
        self.end_headers()

    def do_POST(self):
        self._send_response(200)

    def do_PUT(self):
        self._send_response(200)


def get_sys_path():
    import sys

    return sys.path


def get_sys_import_cache(dir):
    import sys

    try:
        return sys.path_importer_cache[dir]
    except KeyError:
        return "Not found"


def run_python_udf_on_worker(sc, cmd):
    return sc.parallelize(range(0, 1)).map(lambda x: cmd()).collect()


def get_worker_sys_path(sc):
    sys_path = sc.parallelize(range(0, 1)).map(lambda x: get_sys_path()).collect()
    # sys_path[0] is the first return value from parallelize,
    # which in this case is the only entry we care about
    return sys_path[0]


def get_worker_sys_import_cache(sc, notebook_dir):
    import_cache = (
        sc.parallelize(range(0, 1)).map(lambda x: get_sys_import_cache(notebook_dir)).collect()
    )
    # import_cache[0] is the first return value from parallelize,
    # which in this case is the only entry we care about
    return import_cache[0]


class WsfsTestBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # pick a random port
        cls.dummy_server = HTTPServer(("localhost", 0), WsfsDummyServerHandler)

        cls.t = threading.Thread(target=cls.dummy_server.serve_forever)
        cls.t.daemon = True
        cls.t.start()

    def setUp(self, enable_files_in_workspace=False):
        class_name = self.__class__.__name__

        # a test prefix so we can pre-create the directory path, since
        # wsfs won't be actually running
        os.environ["WSFS_TEST_DIR"] = "/tmp"

        conf = SparkConf().set("spark.databricks.wsfsRpcPort", self.dummy_server.server_port)

        if enable_files_in_workspace:
            conf.set("spark.databricks.wsfs.workspacePrivatePreview", "true")
        conf.set("spark.databricks.enableWsfs", "true")
        conf.set("spark.master", "local[1]")
        conf.set("spark.app.name", class_name)

        self.sc = SparkContext.getOrCreate(conf=conf)
        self.sc.setLocalProperty("spark.databricks.api.url", "http://localhost")

    def setupFile(self, folder):
        # Create a temp file and the python process should be running in the dir of the temp file
        os.makedirs("/tmp/Workspace" + folder, exist_ok=True)
        self.test_file = tempfile.NamedTemporaryFile(dir="/tmp/Workspace" + folder)
        self.test_file_name = os.path.basename(self.test_file.name)
        self.sc.setLocalProperty(
            "spark.databricks.notebook.path", folder + "/" + self.test_file_name
        )


class ExecutorPythonPathAndDirectoryTest(WsfsTestBase):
    def test_executor_python_path(self):
        self.setupFile("/Repos/user@email/repo_name/dir")
        sys_path = get_worker_sys_path(self.sc)
        try:
            i = sys_path.index("/tmp/Workspace/Repos/user@email/repo_name/dir")
        except ValueError:
            self.fail(
                "%s did not contain %s"
                % (sys_path, "/tmp/Workspace/Repos/user@email/repo_name/dir")
            )
        self.assertTrue(len(sys_path) > i + 1, sys_path)
        self.assertEquals(sys_path[i + 1], "/tmp/Workspace/Repos/user@email/repo_name")

    def test_executor_python_path_for_internal_repo(self):
        self.setupFile("/Repos/.internal/gitUrl/repo_name/dir")
        sys_path = get_worker_sys_path(self.sc)
        try:
            i = sys_path.index("/tmp/Workspace/Repos/.internal/gitUrl/repo_name/dir")
        except ValueError:
            self.fail(
                "%s did not contain %s"
                % (sys_path, "/tmp/Workspace/Repos/.internal/gitUrl/repo_name/dir")
            )
        self.assertTrue(len(sys_path) > i + 1, sys_path)
        self.assertEquals(sys_path[i + 1], "/tmp/Workspace/Repos/.internal/gitUrl/repo_name")

    def test_executor_python_path_is_repo_path(self):
        self.setupFile("/Repos/user@email/repo_name")
        sys_path = get_worker_sys_path(self.sc)
        from collections import Counter

        count = Counter(sys_path)["/tmp/Workspace/Repos/user@email/repo_name"]
        self.assertEquals(count, 1, sys_path)

    def test_executor_python_process_directory(self):
        self.setupFile("/Repos/foo")

        self.test_file.write(b"testwsfs")
        self.test_file.flush()

        # Check the directory of the python process
        cwd = self.sc.parallelize(range(0, 1)).map(lambda x: os.getcwd()).collect()
        self.assertEqual(["/tmp/Workspace/Repos/foo"], cwd)

        # Check if the python process can read a file under the directory
        file_name = os.path.basename(self.test_file.name)
        content = (
            self.sc.parallelize(range(0, 1)).map(lambda x: open(file_name, "r").read()).collect()
        )
        self.assertEqual(["testwsfs"], content)

    # we ran a non-projects notebook, we should not have added the dir to
    # python path or changed the cwd
    def test_executor_non_repos(self):
        self.setupFile("/Users/foo")
        sys_path = get_worker_sys_path(self.sc)
        for entry in sys_path[0]:
            self.assertFalse("/Users" in entry)

        cwd = self.sc.parallelize(range(0, 1)).map(lambda x: os.getcwd()).collect()
        self.assertFalse("/Users" in cwd[0], cwd)

    def test_change_dir_fails_closed(self):
        # The directory for this notebook does not exist so changing dirs should fail
        self.sc.setLocalProperty(
            "spark.databricks.notebook.path", "/Repos/user@email/repo_name/bad_dir/notebook"
        )

        cwd = self.sc.parallelize(range(0, 1)).map(lambda x: os.getcwd()).collect()
        self.assertEqual(1, len(cwd))
        self.assertFalse("/tmp/Workspace/Repos" in cwd[0], cwd)

        # The python path should still be updated even if we fail to change dirs
        sys_path = get_worker_sys_path(self.sc)
        try:
            i = sys_path.index("/tmp/Workspace/Repos/user@email/repo_name/bad_dir")
        except ValueError:
            self.fail(
                "%s did not contain %s"
                % (sys_path, "/tmp/Workspace/Repos/user@email/repo_name/bad_dir")
            )
        self.assertTrue(len(sys_path) > i + 1, sys_path)
        self.assertEquals(sys_path[i + 1], "/tmp/Workspace/Repos/user@email/repo_name")

    def tearDown(self):
        self.sc.stop()

    @classmethod
    def tearDownClass(cls):
        cls.dummy_server.shutdown()
        cls.t.join()


class FilesInWorkspaceTest(WsfsTestBase):
    def setUp(self):
        super().setUp(enable_files_in_workspace=True)

    def test_executor_non_repos(self):
        self.setupFile("/Users/foo")
        sys_path = get_worker_sys_path(self.sc)
        try:
            i = sys_path.index("/tmp/Workspace/Users/foo")
        except ValueError:
            self.fail("%s did not contain %s" % (sys_path, "/tmp/Workspace/Users/foo"))
        self.assertEquals(len(sys_path) - 1, i)
        self.assertTrue("/WSFS_NOTEBOOK_DIR" not in sys_path)

        cwd = self.sc.parallelize(range(0, 1)).map(lambda x: os.getcwd()).collect()
        # shouldn't change cwd for files in workspace
        self.assertTrue("/Workspace" not in cwd[0])

        # Should instantiate custom import-hook for WSFS.
        sys_path_cache_item = get_worker_sys_import_cache(self.sc, "/tmp/Workspace/Users/foo")
        self.assertTrue(isinstance(sys_path_cache_item, WsfsImportHook))


if __name__ == "__main__":
    import unittest
    from pyspark.tests.test_wsfsdir import *  # noqa: F401

    try:
        import xmlrunner  # type: ignore[import]

        testRunner = xmlrunner.XMLTestRunner(output="target/test-reports", verbosity=2)
    except ImportError:
        testRunner = None
    unittest.main(testRunner=testRunner, verbosity=2)
